# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

try:
    from twisted.internet import glib2reactor
    glib2reactor.install()
except AssertionError:
    # already installed...
    pass

import pkg_resources
import sys, time
import gettext
import locale
import os
import __builtin__
import gobject

try:
    from pgm import render
except ImportError:
    url = "https://core.fluendo.com/pigment/trac/"
    print "You need to install Pigment, check it out at %s" % url
    sys.exit(0)
else:
    # Pigment 0.1.0 has no __version__
    if not hasattr(render, '__version__'):
        render.__version__ = (0, 1, 0)

    required = (0, 1, 5)
    if render.__version__ < required:
        version_str = '.'.join([str(i) for i in required])
        print "Pigment version %s is required, please upgrade." % version_str
        sys.exit(0)
        
from elisa.core import __version__, config, log, menu, common
from elisa.core.utils import exception_hook
from elisa.core.player import player
from elisa.core import skin_manager
from elisa.core import task_manager
from elisa.core.plugin_manager import PluginManager
from elisa.core.media_manager import MediaManager
from elisa.core.signals.application import application_started
from elisa.core.plugin import ExtensionPoint
from elisa.core.interfaces import IUserInputPlugin
from elisa.extern import msgfmt
from elisa.core.update_checker import UpdateChecker
from elisa.core import thumbnail_manager

from pgm.layout import window
from pgm import render
from pgm import log as pgm_log

from twisted.spread import pb
from twisted.internet import reactor, task, threads
from twisted.python import usage

try:
    #from coherence.base import Coherence
    import coherence
except Exception, coherence_error:
    coherence = None
else:
    coherence_error = None
    
WINDOW_TITLE = 'Fluendo Media Center'

class Application(pb.Root):
    """
    Main box application class
    """

    input_plugins = ExtensionPoint(IUserInputPlugin)

    def __init__(self, config_filename=None):
        self.first_run = False
        self.load_localization()
        
        common._set_application(self)
        self.logger = log.Logger()

        self._loop = None
        self._win_update_fd = None
        self._player = None
        self._window = None
        self._skin = None
        self._media_manager = None
        self._config = None
        self._config_filename = config_filename
        
        self.set_config(config_filename)
        self.set_exception_hook()
        
        self.load_profiler()
        
    def set_key_actions(self):
        pgmrender = render.get_renderer()
        self.key_actions = {}
        general_section = self.get_config().get_section('general')

        for name, value in general_section.iteritems():
            if name.endswith('key'):
                pgm_key = getattr(pgmrender, 'KEY_%s' % value)
                self.key_actions[pgm_key] = name[:-4]

    def load_profiler(self):
        enable = int(self.get_config().get_option('enable_profiling',
                                                  default='0'))
        if enable:
            # TODO: import profiling modules and set things ready
            pass
        
        
    def load_localization(self):
        def no_localization():
            # define function _ as identity
            __builtin__.__dict__['_'] = lambda text: text
            
        # retrieve the current locale
        current_locale = locale.getdefaultlocale()[0]
        if current_locale:
            current_language = current_locale[:2]
        else:
            no_localization()
            return

        if sys.platform == "win32": #Cannot use resource_filename with py2exe
            locale_dir = "elisa/core/i18n"
        else:
            locale_dir = pkg_resources.resource_filename(Application.__module__, 'i18n')
        locale_src = os.path.join(locale_dir, current_language, "LC_MESSAGES",
                                  "elisa.po")
        locale_dest = '%s.mo' % os.path.splitext(locale_src)[0]
        
        # check for the existence of a translation file (.po)
        if os.path.exists(locale_src):
            # compile the .po for the current language if necessary
            if not os.path.exists(locale_dest):
                msgfmt.make(locale_src, locale_dest)
            else:
                src_mtime = os.path.getmtime(locale_src)
                dest_mtime = os.path.getmtime(locale_dest)
                if src_mtime > dest_mtime:
                    msgfmt.make(locale_src, locale_dest)
                    
        if os.path.exists(locale_dest):
            translation = gettext.translation("elisa", localedir=locale_dir,
                                              languages=[current_language])
            translation.install()
        else:
            no_localization()

    def setup(self):
        self._root_menuitem = menu.MenuTree()
        
        self.load_renderer()

        self.set_key_actions()

        self._task_manager = task_manager.TaskManager()
        
        config = self.get_config()

        media_manager = MediaManager(config)
        self.set_media_manager(media_manager)
        media_manager.start()
        
        plugin_manager = PluginManager(config)
        common._set_plugin_manager(plugin_manager)
        plugin_manager.load_plugins()
        
        skin_name = config.get_option('skin', default='elisa')

        self.create_menu()

        if skin_name:
            self.create_window()
            self._skin_manager = skin_manager.SkinManager(skin_name,
                                                          self._window,
                                                          self._root_menuitem)
            self._skin = self._skin_manager.get_skin()
            
        self._player = self._load_player()
        self._thumbnail_manager = thumbnail_manager.ThumbnailManager()
        
        self._coherence = self._load_coherence()

        
    def load_renderer(self):
        config = self.get_config()
        renderer_name = config.get_option('renderer')
        render.init(renderer_name)
        
    def create_window(self):
        config = self.get_config()
        size = config.get_option('resolution', default=None)
        fs_size = config.get_option('fullscreen_resolution', default=None)   
        start_fs = config.get_option('start_fullscreen', default=0)
                    
        self._window = window.Window(size=size,
                                     title=WINDOW_TITLE,
                                     fullscreen=start_fs,
                                     scale_mode=window.SCREEN_ADJUST)

        if start_fs:
            # hide cursor
            pgmrender = render.get_renderer()
            self._window.cursor = pgmrender.CURSOR_NONE
            self._window.set_fullscreen(True, fs_size)

        self._win_update_fd = gobject.timeout_add(15, self._window.update)
        
    def _load_player(self):
        queue = []
        active_queue = self.get_config().get_option('active_queue', default='')
        if active_queue:
            plugin_manager = common.get_plugin_manager()
            plugin_manager.load_plugins_for_entry_point('elisa.plugins.queue',
                                                        [active_queue])
            queue = plugin_manager.get_plugin_with_name(active_queue)
            queue.post_init()

        return player.Player(queue=queue)

    def _load_coherence(self):
        coherence_instance = None
        required_version = (0, 2, 1)
        if coherence:
            try:
                from coherence import __version_info__
            except ImportError:
                __version_info__ = (0, 1, 0)

            from coherence.base import Coherence
            if __version_info__ < required_version:
                required = '.'.join([str(i) for i in required_version])
                found = '.'.join([str(i) for i in __version_info__])
                self.logger.info("Coherence %s required. %s found. "\
                                 " Please upgrade" % (required, found))
            else:
                default = {'logmode':'warning',
                           'controlpoint': 'yes',
                           'plugins':{}}
                coherence_config = self.get_config().get_section('coherence',
                                                                 default)
                coherence_config['plugins'] = {}
                try:
                    coherence_instance = Coherence(coherence_config)
                except Exception, coherence_exception:
                    self.logger.info("Coherence failed to start")
                    self.logger.info("Error: %s" % coherence_exception)
        else:
            self.logger.info("Coherence UPnP A/V framework not found.")
            self.logger.info("Error: %s" % coherence_error)
        return coherence_instance
    
    def get_window(self):
        return self._window
        
    def get_player(self):
        return self._player

    remote_get_player = get_player

    def get_coherence(self):
        return self._coherence

    def get_skin_manager(self):
        return self._skin_manager

    def get_thumbnail_manager(self):
        return self._thumbnail_manager

    def get_task_manager(self):
        return self._task_manager

    def set_media_manager(self, manager):
        self._media_manager = manager

    def get_media_manager(self):
        return self._media_manager

    remote_get_media_manager = get_media_manager
    remote_get_plugin_manager = lambda self: common.get_plugin_manager()

    # backward compatibility
    remote_get_cache_manager = get_media_manager
    
    def get_skin(self):
        return self._skin

    def set_config(self, filename):
        try:
            self._config = config.Config(filename)
        except Exception, exc:
            self.close()

    def get_config(self):
        """ Return the application's config which is a `config.Config`
        instance.
        """
        return self._config

    def load_image_uri(self, uri, in_slideshow=False):
        self.get_skin().display_image(uri, in_slideshow=in_slideshow)

    def load_stream_uri(self, player, uri):
        return self.get_skin().display_stream(player, uri)

    def set_exception_hook(self):
        """ Override the default system exception hook with our
        EmailTracebackHook.
        """
        mail_section = self.get_config().get_section('mail')

        enable = int(mail_section.get('enable', '0'))
        if enable:
            mail_from = mail_section.get('mail_from', 'elisa@localhost')
            mail_to = mail_section.get('mail_to', ['phil@localhost'])
            mail_subject = "Elisa error"
            smtp_server = mail_section.get('smtp_server', 'localhost')

            exception_hook.set_exception_hook(mail_from, mail_to,
                                              mail_subject, smtp_server)

    def create_menu(self):
        """Create menu from plugin list
        """
        plugins = common.get_plugin_manager().get_main_plugins()
        for tree in plugins:
            self._root_menuitem.add_item(tree)
        # TODO: re-enable this when gui.widget.hide() would be working
        # self._root_menuitem.add_item(self._task_manager)
        
    def run(self, loop):
        """ Execute the application. Start main loop.
        """
        self.check_for_updates()
        self._loop = loop
        
        skin = self.get_skin()

        if skin:
            self.get_skin().prepare()
            self._window.sig_key_press_event.connect(self.key_pressed)
            self._window.sig_button_press_event.connect(self.button_pressed)
            self._window.sig_exit_event.connect(lambda _: self.close())
            reactor.callLater(1.0 / 25, self.refresh)

        if skin:
            self._window.main(loop)
            
        application_started.emit()

        if self.first_run and skin:
            msg = (_('Read this page to configure Elisa'),'',
                   'https://core.fluendo.com/elisa/trac/wiki/FirstRun')
            skin.notify(msg, font_size=23)
            for line in msg:
                print line
        
    def restart(self):
        self.close(True)
        
    def check_for_updates(self):
        up = UpdateChecker(self)
        up.start()
        
    def refresh(self):
        additional_events = []        
        for feature, plugin in self.input_plugins.iteritems():
            additional_events.extend(plugin.get_input_events())

        win = self.get_window()
        if win:
            win.push_events(additional_events)
        
	reactor.callLater(1.0 / 25, self.refresh)

    def close(self, restart=False):
        """ Close the application. Good idea to save the configuration
        here, since it's probable it has been updated.
        """

        def skin_unloaded(r):
            player = self.get_player()
            if player:
                player.close()
            manager = common.get_plugin_manager()
            if manager:
                manager.unload_plugins()
            media_manager = self.get_media_manager()
            if media_manager:
                media_manager.stop()
            thumbnail_manager = self.get_thumbnail_manager()
            if thumbnail_manager:
                thumbnail_manager.stop()

            coherence = self.get_coherence()
            if coherence:
                coherence.shutdown()

            config = self.get_config()
            if config:
                config.write()

            gobject.source_remove(self._win_update_fd)

            # reset screen resolution and quit pigment loop
            win = self.get_window()
            if win:
                win.set_fullscreen(False, win.screen_resolution)
                win.main_quit()

            if restart:
                args = sys.argv
                args.insert(0, sys.executable)
                cmd = " ".join(args)
                if reactor.running:
                    reactor.disconnectAll()
                    reactor.stop()
                os.spawnvpe(os.P_WAIT, sys.executable, args, os.environ)
            
            if reactor.running:
                reactor.stop()
            else:
                sys.exit(-1)

        dfr = None
        skin = self.get_skin()
        if skin:
            dfr = skin.unload()
            dfr.addCallback(skin_unloaded)
        else:
            skin_unloaded(None)
        return dfr
    
    def key_pressed(self, key_event):
        pgmrender = render.get_renderer()

        key = key_event.keyval
        if self.key_actions.has_key(key):
            action = self.key_actions.get(key)
            if action and hasattr(self, action):
                getattr(self, action)()

    def button_pressed(self, event):
        pass


class Options(usage.Options):
    optFlags = [['elisa-version', '', 'show elisa version'],
                ['pigment-version', '', 'show pigment version'],
                ]

    def parseArgs(self, config_file=None):
        self['config_file'] = config_file


def start():

    options = Options()
    try:
        options.parseOptions()
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)

    sys_exit = False
    if options['elisa-version']:
        print 'Elisa version %s' % __version__
        sys_exit = True

    # TODO: move pigment specific options to Pigment...
    if options['pigment-version']:
        pgm_version = '.'.join([str(i) for i in render.__version__])
        print 'Pigment version %s' % pgm_version
        sys_exit = True

    if sys_exit:
        sys.exit(0)

    app = Application(options['config_file'])
    app.setup()
    app.run(reactor.loop)

    try:
        reactor.listenTCP(8789, pb.PBServerFactory(app))
    except:
        pass
    reactor.run()
    app.close()
