# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

import os, sys

from elisa.core.menu import MenuItem
from elisa.core.plugin import Plugin
from elisa.core import common, media_uri, log
from elisa.core import menu_manager
from elisa.core.signals.menu import *
from elisa.core.signals.skin import *
from elisa.core.utils.misc import get_resource_path
from elisa.core.utils import image

from pgm import render, widget
from pgm.layout import group
from pgm.form import rectangle

from pgm import log as log_pgm

class BaseSkin(Plugin):

    # relative path to the directory storing the various skins pictures
    pictures_dir = "override_me"

    icon_size = "128x128"
    
    def __init__(self, window, root_menu_item):    
        Plugin.__init__(self)
        
        self._window = window
        self._menu_manager = menu_manager.MenuManager(root_menu_item)
        self.key_actions = {}
        self._wallpaper_image_path = None
        self._background_widget = None
        self._is_menu_visible = True

    def get_window(self):
        return self._window
        
    def get_menu_manager(self):
        return self._menu_manager

    def set_wallpaper(self, image_path):
        if self._background_widget:
            pgmrender = render.get_renderer()
            self._background_widget.aspect_style = pgmrender.LAYOUT_FILLED

            self._wallpaper_image_path = image_path
            img = image.Image(image_path)
            self._background_widget.media = img
            self._background_widget.tz = -1
            
    def get_wallpaper(self):
        return self._wallpaper_image_path
        
    def set_background_media(self, media):
        """
        Display a media in the background
        """

        if self._background_widget:
            pgmrender = render.get_renderer()
            self._background_widget.aspect_style = pgmrender.LAYOUT_SCALED
            self._background_widget.media = media

    def rotate_image(self, uri, angle):
        img = image.Image(uri)
        img.rotate(angle)
        if self._background_widget:
            media = self._background_widget.media
            # FIXME: should rotate background if necessary
            if isinstance(media, image.Image) and False:
                media.rotate(angle)
                self._background_widget.media = media

    @classmethod
    def get_picture_path(cls, relative_path, icon_size=None):
        """
        """
        if not icon_size:
            icon_size = cls.icon_size
        path = None
        if relative_path:
            if not os.path.isfile(relative_path):
                
                path = os.path.join(cls.pictures_dir, icon_size, relative_path)
                path = get_resource_path(cls.__module__, path)
                path = path.replace(os.path.sep, '/')
            else:
                path = relative_path
        return path
        
    def is_icon(self, picture_path):
        basename = os.path.basename(picture_path)
        return picture_path == self.get_picture_path(basename)
        
    def prepare(self):
        # Key press handling on the whole window
        self._window.sig_key_press_event.connect(self.keypress_cb)

        self._window.sig_resize_event.connect(self.resize_cb)

        # Root menu construction
        self.get_menu_manager().build_root_menu()
        self.set_key_actions()

        self._setup_background_widget()
        
    def _setup_background_widget(self):
        pgmrender = render.get_renderer()

        win_size = self.get_window().canvas_size
        self._grp = group.Group(size=win_size)
        self._grp.ordering = widget.FAR
        self._window.add(self._grp)
        self._background_widget = rectangle.Rectangle(win_size)
        self._background_widget.ordering = widget.FAR
        self._grp.add(self._background_widget)
        self._background_widget.tz = -1
        
    def set_key_actions(self):
        self.key_actions = {}
        pgmrender = render.get_renderer()
        config = common.get_application().get_config()
        
        for opt_name, value in self.get_config().iteritems():
            if not opt_name.endswith('_key'):
                continue
            opt_name = opt_name[:-4]
            value = 'KEY_%s' % value
            try:
                key_value = getattr(pgmrender, value)
            except AttributeError:
                msg = "Skipping configuration of %r keybinding" % value
                self.logger.info(msg)
                continue
            self.key_actions[key_value] = opt_name
        
    def keypress_cb(self, key_event):
        print key_event
        key = key_event.keyval
        if self.key_actions.has_key(key):
            action = self.key_actions.get(key)
            if action and hasattr(self, action):
                getattr(self, action)()

    def resize_cb(self, resize_event):
        pass

    def add_widget(self, widget):
        self._grp.add(widget)

    def remove_widget(self, widget):
        self._grp.remove(widget)

    # Key actions
            
    def activate(self):
        manager = self.get_menu_manager()
        activated = False
        if not manager.is_locked():
            activated = manager.activate_current_menu_item()            
        return activated

    def move_right(self):
        manager = self.get_menu_manager()
        if manager.is_locked():
            return False
        return manager.go_next_menu_item()

    def move_left(self):
        manager = self.get_menu_manager()
        if manager.is_locked():
            return False
        return manager.go_previous_menu_item()

    def move_down(self):
        manager = self.get_menu_manager()
        if manager.is_locked():
            return False
        return True

    def move_up(self):
        manager = self.get_menu_manager()
        if manager.is_locked():
            return False
        
        menu_item = manager.get_current_menu_item()
        if menu_item.get_parent() != manager.get_root_item():
            return manager.go_up_level()
        else:
            return False
        
        return True
                
    def toggle_fullscreen(self):
        is_fullscreen = self._window.get_fullscreen()
        pgmrender = render.get_renderer()

        if is_fullscreen:
            # show cursor
            self._window.cursor = pgmrender.CURSOR_INHERIT
        else:
            # hide cursor
            self._window.cursor = pgmrender.CURSOR_NONE
            
        fs_size = None
        if not is_fullscreen:
            config = common.get_application().get_config()
            fs_size = config.get_option('fullscreen_resolution', default=None)
            
        self._window.set_fullscreen(not is_fullscreen, fs_size)
        self._window.regenerate_widgets()
        
    def toggle_menu(self):
        if self._is_menu_visible:
            self.hide_menu()
        else:
            self.show_menu()

    def menu_visible(self):
        return self._is_menu_visible

    def show_menu(self):
        self._is_menu_visible = True

    def hide_menu(self):
        self._is_menu_visible = False

    def notify(self, message):
        pass
