# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from sets import Set
import inspect
import elisa.core
import os
import platform
from twisted.spread import pb

class Interface:
    """

    """

class InterfaceOmission(Exception):
    """
    Raised when an un-implemented method is found in the plugin
    """

class UnsupportedPlatform(Exception):
    """
    Raised when the plugin does not support user's platform
    """

class InterfaceChecker(type):
    """
    Validates the presence of required attributes
    """

    def __new__(cls, classname, bases, classdict):
        """ called when the Plugin class is created. We check:

            - the plugin __implements__ some interface
            - the plugin methods are all referenced in the interface
        """
        obj = type.__new__(cls, classname, bases, classdict)
        interfaces = classdict.get('__implements__')

        if type(interfaces) not in (list, tuple):
            interfaces = [interfaces,]

        platforms = classdict.get('__platforms__')    
        if platforms:
            user_platform = [os.name, platform.system()]
            if not ((user_platform[0] in platforms) or (user_platform[1] in platforms)):
                error = "Platform not supported for %r; supported are: %r"
                raise UnsupportedPlatform, error % (obj, platforms)

        interfaces = [ iface for iface in interfaces if iface ]
        for interface in interfaces:
            defined = Set(dir(obj))
            required = Set(dir(interface))
            if not required.issubset(defined):
                missing = list(required - defined)
                error = "Not implemented methods from %s in %r : %r"
                raise InterfaceOmission, error % (interface.__name__,
                                                  obj, missing)

        # safety check
        class_names = ['Component', 'SignalEmitter', 'Plugin'] + \
                      elisa.core.BASE_PLUGINS
        if classname not in class_names:
            assert obj.implements(Interface), \
                   "You need to implement at least Interface in %r" % classname

        # bind method documentation from interfaces to implementations
        for interface in interfaces:
            for attr_name in dir(interface):
                if attr_name in ('__module__',):
                    continue
                attr = getattr(interface, attr_name)
                if hasattr(attr, '__doc__'):
                    method = getattr(obj, attr_name)
                    if callable(method) and hasattr(method, 'im_func'):
                        method.im_func.func_doc = attr.__doc__

                # optionally export the method to PB
                private = hasattr(attr, 'private') and attr.private
                if not private:
                    setattr(obj, 'remote_%s' % attr_name, method)
                            
        return obj

class Component(object, pb.Referenceable):
    __metaclass__ = InterfaceChecker

    @classmethod
    def implements(cls, interface):
        """
        Check the class implements the given interface which is searched
        "deeply" in the __implements__ list.
        """

        assert hasattr(cls, '__implements__'), \
               "Missing __implements__ in %r" % cls

        found = False
        interfaces = cls.__implements__
        if type(interfaces) not in (list, tuple):
            interfaces = [interfaces,]
        for i in interfaces:
            mro = inspect.getmro(i)
            if interface in mro:
                found = True
                break
        return found

if __name__ == '__main__':

    class ITest(Interface):

        def foo():
            """ bar """
            

    class TestComponent(Component):
        __implements__ = ITest

        def foo(self):
            print 'boo'

    c = TestComponent()
    print c.foo.__doc__
