# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pysqlite2 import dbapi2 as sqlite
from elisa.core import media_uri, log
from elisa.core.utils import misc
from elisa.core import common
import os
import time

VERSION = 1

CREATE_SQL = """
PRAGMA default_synchronous = OFF;

CREATE TABLE meta (
    version INTEGER UNIQUE
);

CREATE TABLE source (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    uri TEXT NOT NULL,
    short_name TEXT,
    available INT DEFAULT 1
);

CREATE UNIQUE INDEX INDX_SOURCE_ID ON source(id);

CREATE TABLE media (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    uri TEXT NOT NULL,
    short_name TEXT,
    media_node_id INT DEFAULT 0,
    source_id INT DEFAULT 0,
    format TEXT DEFAULT '',
    typ TEXT DEFAULT '',
    deleted INT DEFAULT 0,
    rating INT DEFAULT 50,
    date_added DATETIME,
    last_played DATETIME DEFAULT NULL,
    updated INT DEFAULT 0
);

CREATE TABLE audio_media (
    media_id INT UNIQUE,
    artist TEXT DEFAULT 'unknown artist',
    album TEXT DEFAULT 'unknown album',
    song TEXT DEFAULT '',
    track INT DEFAULT 0,
    cover_uri TEXT DEFAULT ''
);

CREATE TABLE video_media (
    media_id INT UNIQUE
);

CREATE TABLE image_media (
    media_id INT UNIQUE
);

CREATE INDEX INDX_MEDIA_URI ON media(uri);
CREATE INDEX INDX_MEDIA_SOURCEID ON media(source_id);
CREATE INDEX INDX_MEDIA_MD_ID ON media(media_node_id);
CREATE INDEX INDX_MEDIA_ARTIST ON audio_media(artist);

CREATE INDEX INDX_AUDIO_MEDIA_ID ON audio_media(media_id);
"""


unicode_to_db = misc.fs_encode

if hasattr(sqlite,'Row'):
    Row_parent = sqlite.Row
else:
    Row_parent = dict

MEDIA_COLUMNS={}

class Row(Row_parent):

    _cache = {}

    def __getattr__(self, attr):
        try:
            value = self._cache.get(attr, self[attr])
        except IndexError:
            value = ''
        if type(value) == unicode:
            value = value.decode('string_escape')
            value = misc.fs_decode(value)
        return value

    def has_key(self, key):
        try:
            self[key]
        except:
            result = False
        else:
            result = True
        return result

    def __setattr__(self, key, value):
        self._cache[key] = value

class DummyRow(dict):

    _cache = {}

    def __getattr__(self, attr):
        try:
            value = self._cache.get(attr, self[attr])
        except IndexError:
            value = ''
        if type(value) == unicode:
            value = value.decode('string_escape')
            value = misc.fs_decode(value)
        return value

    def has_key(self, key):
        try:
            self[key]
        except:
            result = False
        else:
            result = True
        return result

    def __setattr__(self, key, value):
        self._cache[key] = value

class MediaDB:

    def __init__(self, path, reset=False):
        if not os.path.exists(path):
            reset = True
        self.logger = log.Logger()
        self.counter = 0
        self._db_file = path
        self._media_columns = {}
        self._connect()
        self.init_database(reset)

    def _connect(self):
        self._con = sqlite.connect(database=self._db_file,timeout=3,
                                   check_same_thread=True)
        if hasattr(self._con,'row_factory'):
            self._con.row_factory = Row
            self._build_rows = False
        else:
            self._build_rows = True
        self._con.isolation_level = None
        
    def init_database(self, reset=False):
        if reset:
            for table_name in ("source", "media","audio_media", "video_media",
                               "picture_media", "meta"):
                self.execute("drop table %s" % table_name, hide_error=True)
            for sql in CREATE_SQL.split(';'):
                if sql:
                    self.execute(sql, True, True)
            
            self.execute("insert into meta(version) values(%s)" % VERSION)
        else:
            # handle db schema upgrades
            
            # [0.1.3 -> 0.1.4] check if "meta" table exists
            result = self.execute("PRAGMA table_info(meta)")
            if not result:
                # Hummer-mode-on: drop old db and restart from scratch
                self.logger.info("Database wiped out, regenerating it from scratch")
                os.unlink(self._db_file)
                self._connect()
                self.init_database(True)

    def save_changes(self):        
        try:
            self._con.commit()
        except Exception, ex:
            print 'Commit failed, %s' % ex
            
    def _query(self, request):
        if self._build_rows:
            cursor = self._con.cursor()
            cursor.execute(request)
            if not cursor.description:
                result = []
            else:
                headers = [ d[0] for d in cursor.description ]
                result = [ Row(dict(zip(headers, row)))
                           for row in cursor.fetchall() ]
            cursor.close()
        else:
            result = self._con.execute(request).fetchall()
        return result

    def execute(self, request, hide_error=False, do_commit=False, log=False):
        #log = True
        t0 = time.time()
        if log:
            self.logger.info("SQL: %s" % request)
        result = None
        if hide_error:
            try:
                result = self._query(request)
            except Exception,exc:
                pass
                
        else:
            try:
                result = self._query(request)
            except Exception, exc:
                self.logger.info("Exception occurred: %s" % exc)
                self.logger.info("SQL was: %s" % request)
                

        if result:
            self.counter += 1

        if do_commit or self.counter >= 50:
            self.save_changes()
            
        if log:
            delta = time.time() - t0
            self.logger.info("SQL request took %s seconds" % delta)
            
        return result

    def close(self):
        self.logger.info('Closing DB connection ...  %s' % id(self))
        #self.save_changes()
        #self._con.thread.join(timeout=1.)
        #self._con.close()
        
    def _clean_uri(self, uri):
        uri = str(uri)
        if uri.endswith('*'):
            uri = uri[:-1]
        uri = str(media_uri.Uri(location=uri))
        return uri


    def _kw_extract(self, kw):
        values = []
        spec = []

        for col_name, value in kw.iteritems():
            if value == None:
                continue
            spec.append(col_name)
            if type(value) == unicode:
                value = unicode_to_db(value)
            elif type(value) == long:
                value = int(value)
            value_rep = repr(value)
            values.append(value_rep)
            
        spec = ','.join(spec)
        values = ','.join(values)
        return (spec, values)

    def _guess_format(self, metadata):
        
        # lazily introspect the *_media tables and store column names in a hash
        global MEDIA_COLUMNS
        if not MEDIA_COLUMNS:
            for fmt in ('audio','video','picture'):
                table_name = "%s_media" % fmt
                result = self.execute("PRAGMA table_info(%s)" % table_name)
                col_names = set([ r.name for r in result ])
                MEDIA_COLUMNS[fmt] = col_names
                
        format = ''
        keys = set(metadata.keys())
        for fmt, col_names in MEDIA_COLUMNS.iteritems():
            if keys.issubset(col_names):
                format = fmt
                break
        return format

    ##################################################################
    # API
    ##################################################################

    def prepare_source_for_update(self, source):
        self.logger.info('Preparing %s' % source.uri)
        req = 'update media set updated=0 where source_id=%s and deleted=0' % source.id
        self.execute(req)

    def hide_un_updated_medias_for_source(self, source):
        req = 'select count(id) as c from media where updated=0 and source_id=%s' % source.id
        count = self.execute(req)[0].c
        if count:
            self.logger.info("Deleting %s medias from db" % count)
##         req = 'select * from media where updated=0 and source_id=%s' % source.id
##         for row in self.execute(req):
##             print 'deleting %s' % row.uri
        
            req = 'update media set updated=1, deleted=1 where source_id=%s and updated=0' % source.id
            self.execute(req)

    def get_medias(self, source=None):
        req = 'select * from media where'
        if source:
            where = 'source_id=%s' % source.id
        else:
            where = 'media_node_id=-1'
        req = "%s %s and deleted=0" % (req, where)
        return self.execute(req)

    def get_files_count_for_source(self, source):
        request = "select count(*) as c from media where typ='file' and source_id=%s" % source.id
        count = self.execute(request)[0].c
        return count

    def add_source(self, uri, short_name):
        uri = self._clean_uri(uri)
        uri = unicode_to_db(uri)
        request = "insert into source(uri, short_name) values(%r, %r)" % (uri,
                                                                          short_name)
        self.execute(request)
        source = self.get_source_for_uri(uri)
        now = 'current_timestamp'
        request = "insert into media(uri, updated, date_added, media_node_id, short_name, source_id) values (%r, 1, %s, -1, %r, %s)" % (uri, now, short_name, source.id)
        self.execute(request)
        return source

    def updated_source(self, source):
        request = "update media set updated=1 where source_id=%s and media_node_id=-1" % source.id
        self.execute(request)

    def hide_source(self, source):
        request = "update source set available=0 where id=%s" % source.id
        self.execute(request)

    def show_source(self, source):
        request = "update source set available=1 where id=%s" % source.id
        self.execute(request)
        
    def add_media_node(self, uri, short_name, **extras):
        node = self.get_media_for_uri(uri)
        do_insert = True
        do_update = False
        media_node_id = extras.get('media_node_id')
        if node:
            if media_node_id and node.media_node_id == media_node_id:
                do_insert = False
            if node.deleted and node.artist != 'unknown artist':
                do_update = True

        metadata = extras.get('metadata',{})
        if 'metadata' in extras:
            del extras['metadata']

        if do_insert:
            extras_spc, extras_val = self._kw_extract(extras)
                                  
            uri = unicode_to_db(uri)
            if short_name:
                short_name = unicode_to_db(short_name)
            else:
                short_name = ''

            now = 'current_timestamp'
            req = "insert into media(uri, updated, date_added, short_name,%s) values (%r,1, %s, %r,%s)"
            req = req % (extras_spc, uri, now, short_name, extras_val)
            self.execute(req)
            node = self.get_media_for_uri(uri)

            # add new row in specialized table
            format = extras.get('format')
            if format:
                spec, values = self._kw_extract(metadata)
                if spec and values:
                    req = "insert into %s_media(media_id,%s) values (%s,%s)"
                    req = req % (format, spec, node.id, values)
                else:
                    req = "insert into %s_media(media_id) values (%s)" 
                    req = req % (format, node.id)
                self.execute(req)
                
        elif do_update:
            extras.update(deleted=0, updated=1)
            self.update(node, **extras)
        return do_insert, do_update, node

    def del_media_node(self, node):
        if not node.deleted:
            self.logger.info("Dropping Media %s from DB" % node.uri)
            req = "update media set deleted=1 where id=%s" % node.id
            self.execute(req, hide_error=False, do_commit=True)
            
    def get_parent(self, node):
        parent = None
        if node.media_node_id != -1:
            parent = self.get_media_with_id(node.media_node_id)
        return parent

    def get_source_for_uri(self, uri):
        request = "select * from source where uri=%r"
        uri = misc.fs_encode(uri)
        uri = str(media_uri.Uri(location=str(uri)))
        rows = self.execute(request % uri)
        if rows:
            return rows[0]

    def get_media_for_uri(self, uri):
        request = "select * from media where uri=%r"
        uri = misc.fs_encode(uri)
        uri = str(media_uri.Uri(location=str(uri)))
        rows = self.execute(request % uri)
        media = None
        if rows:
            media_row = rows[0]
            if media_row.format:
                req = "select m.*, s.* from media m, %s_media s where id=%s and id=media_id"
                req = req % (media_row.format, media_row.id)
                result = self.execute(req)
                if result:
                    media = result[0]
                else:
                    media = media_row
            else:
                media = media_row
        return media

    def get_nodes(self, media_or_node):
        if (not media_or_node) or media_or_node.deleted:
            return []
        key = "media_node_id"
        request = "select * from media where %s=%s and deleted=0" % (key,
                                                                     media_or_node.id)
        nodes = self.execute(request)
        nodes.sort(key=lambda n: n.short_name)
        return nodes

    def get_media_with_id(self, media_id):
        media = None
        request = "select * from media where id=%s limit 1" % media_id
        result = self.execute(request)
        if result:
            media = result[0]
        return media
    
    def node_is_dir(self, node):
        return node.has_key('typ') and node.typ == 'directory'

    def node_is_file(self, node):
        return node.has_key('typ') and node.typ != 'directory'

    def node_is_empty(self, media_node, for_plugin):

        def _empty(node):
            is_empty = True

            or_format = ["format=%r" % format for format in for_plugin.provides]
            if or_format:
                request = "select id from media where media_node_id=%s and %s limit 1"
                request = request % (node.id, ' or '.join(or_format))
                result = self.execute(request)
                if result:
                    is_empty = False
            return is_empty

        empty = _empty(media_node)
        if empty:
            for node in self.get_nodes(media_node):
                if self.node_is_dir(node):
                    empty = _empty(node)
                    if not empty:
                        break
        return empty
        
    def node_is_empty_for_all(self, media_node):
        plugin_mgr = common.get_plugin_manager()
        plugins = plugin_mgr.get_main_plugins()
        for plugin in plugins:
            if not self.node_is_empty(media_node, plugin):
                return False
        return True

    def update(self, node, **kw):
        table = "media"
        parsed = {}
        for k,v in kw.iteritems():
            if type(v) == unicode:
                v = misc.fs_encode(v)
            if type(v) == long:
                v = int(v)
            if v is not None:
                parsed[k] = v
        values = ', '.join(["%s=%r" % (k,v) for k,v in parsed.iteritems()])
        if values:
            req = "update %s set %s where id=%s" % (table, values, node.id)
            self.execute(req)

    def update_metadata(self, node, parent, **metadata):
        format = ''
        if not parent.format:
            if node:
                format = node.format
            else:
                format = self._guess_format(metadata)
        else:
            format = parent.format

        if format:
            table = "%s_media" % format

            req = "select count(media_id) as c, %s.* from %s, media where media_id=%s and id=media_id"
            result = self.execute(req % (table, table, parent.id))
            if not result or result[0].c == 0:
                spec, values = self._kw_extract(metadata)
                if spec:
                    req = "insert into %s(media_id,%s) values(%s,%s)"
                    req = req % (table, spec, parent.id, values)
                else:
                    req = "insert into %s(media_id) values(%s)"
                    req = req % (table, parent.id)
                self.execute(req)
            else:
                parent = result[0]
                parsed = {}
                for k,v in metadata.iteritems():
                    if type(v) == unicode:
                        v = misc.fs_encode(v)
                    if type(v) == long:
                        v = int(v)
                    if v is not None:
                        parsed[k] = v
                values = ', '.join(["%s=%r" % (k,v)
                                    for k,v in parsed.iteritems()])
                if self._node_changed(parent, parsed) and values:
                    req = "update %s set %s where media_id=%s" % (table, values,
                                                                  parent.media_id)
                    self.execute(req)
                    for k,v in parsed.iteritems():
                        setattr(parent, k, v)
                        
    def _node_changed(self, node, values):
        changed = False
        for key, value in values.iteritems():
            stored_value = getattr(node, key)
            if type(stored_value) == unicode:
                stored_value = stored_value.encode('utf-8')
            if stored_value != value:
                changed = True
                break
                
        return changed
        
if __name__ == '__main__':
    db = MediaDB('elisa.db')
    print db.get_media_for_uri('file:///home/phil/devel/Elisa/sample_data/movies/')
    print db.get_media_for_uri('file:///home/phil/devel/Elisa/sample_data/pictures/animals/swan.jpg')
##     print db.get_media_for_uri('foo')

##     m = db.add_media('foo','bar')
##     print m
