# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core.utils import misc
import urlparse, urllib
import sys, os, re
import copy

def register_scheme(name):
    """
    Add a scheme (http, smb...) to the list of recognized schemes.

    :parameters:
        `name` : string
            Scheme to add
    """
    if name not in urlparse.uses_netloc:
        urlparse.uses_netloc.append(name)

class Uri:

    """
    TODO
    """


    def __init__(self, location=None, parts=None, quote=False):
        """
        Create an Uri from various representations.

        Representation can be either a string containing the uri or
        the components of the uri stored in a dictionary.

        :parameters:
            `location` : string
                An uri stored in a string, possibly unicode
            `parts` : dictionary
                An uri described by its components (scheme, user,
                password, host, port, path, params, fragment), each of them
                being a string
            `quote` : TODO
                TODO

        :exceptions:
            - `UnicodeDecodeError` If encoding is unicode and decoding
                                    to local filesystem encoding failed
            - `TypeError` If none of location or parts has been provided
        """
        self.quote = quote
        if location:
            location = location.replace(os.path.sep, '/')
            if type(location) == unicode:
                try:
                    location = misc.fs_encode(location)
                except UnicodeDecodeError:
                    pass

            self.location = location
            self.parsed = False
            parts = {'scheme':'',
                     'user':'', 'password':'',
                     'host':'', 'port': '',
                     'path':'', 'params':'',
                     'fragment':''}
            self._update_parts(parts)
            self._parse()
        elif parts:
            self.parsed = True
            self._update_parts(parts)
        else:
            raise TypeError("You need to supply either location or parts hash")
        
    def copy(self):
        """ Return a shallow copy of the Uri instance """
        # FIXME: to remove
        return copy.copy(self)

    def get_label(self):
        """ Return a displayable string designing the Uri """
        location = str(self)
        if location == 'smb:///':
            label = "Samba shares"
        else:
            label = self.parts['params'].get('label')
            if not label:
                path = self.path
                if path.endswith('/'):
                    path = path[:-1]
                label = os.path.basename(path)
        return label

    def get_dirname(self):
        """ Return the directory name in a string

            Examples:
            ssh:///boudiou/truc/com/ -> /boudiou/truc/com
            ssh:///boudiou/truc/com -> /boudiou/truc
        """
        # FIXME: remove
        location = str(self)
        if location == 'smb:///':
            dirname = "/"
        else:
            splitted = self.path.split('/')
            dirname = '/'.join(splitted[:-1])
        return dirname

    def get_parent(self):
        """ Return scheme://get_dirname() """
        scheme = self.scheme
        dirname = self.get_dirname()
        port = self.port
        if port:
            host_port = "%s:%s" % (self.host, port)
        else:
            host_port = self.host
        return "%s://%s%s" % (scheme, host_port, dirname)

    def is_child(self, other_uri):
        """ Tell if a Uri is contained into another one """
        # FIXME: to remove
        path = self.host + self.path
        other_path = other_uri.host + other_uri.path
        result =other_uri.scheme == self.scheme and path.startswith(other_path)
        return result

    def _parse(self):
        """scheme://netloc/path/?parameters#fragment.
           scheme://user:pass@netloc/path/?params#fragment

        """
        parse_result = urlparse.urlparse(self.location)
        scheme, netloc, url, params, query, fragment = parse_result

        # re-parse uri if necessary
        if scheme and scheme != 'http':
            old_scheme = scheme
            match = re.match('.*://(.*)', self.location)
            if match:
                loc = match.groups()[0]
                found_params = re.match('.*\?.*=.*', self.location)
                if scheme == 'file' and not found_params:
                    if sys.platform != 'win32': #FIXME: check for possible misuse
                        loc = urllib.quote(loc)
                parse_result = urlparse.urlparse('http://%s' % loc)
                scheme, netloc, url, params, query, fragment = parse_result
                scheme = old_scheme
        
        user = ''; password = ''; port = ''
        if netloc.find('@') > 0:
            segments = netloc.split('@')
            user_pass = segments[0]
            netloc = segments[1]

            # parse user:password
            if user_pass.find(':') > 0:
                user, password = user_pass.split(':')

        self.safe = True
        if query:
            try:
                params = dict([v.split('=') for v in query.split('&')])
            except:
                self.safe = False
                query = query.replace(',','&')
                try:
                    params = dict([v.split('=') for v in query.split('&')])
                except:
                    import pdb; pdb.set_trace()
            else:
                self.safe = True
            if self.quote:
                for k,v in params.iteritems():
                    params[k] = urllib.unquote(v)
            path = url
        else:
            path = url
            params = {}

        # parse host:port
        if netloc.find(':') > 0:
            host, port = netloc.split(':')
        else:
            host = netloc

        if sys.platform == 'win32' and scheme not in urlparse.uses_netloc:
            host, scheme = scheme, host
        
        if not scheme:
            scheme = 'file'

        if scheme == 'file':
            if host == '.':
                path = "." + path
                host = ''

            path = os.path.abspath(path)
            '''
            if not host and sys.platform == 'win32':
                host = os.getcwd().replace(os.path.sep,'/')
            '''

            if sys.platform == 'win32' and host.find(':') == -1:
                host = '%s:' % host

            path = urllib.unquote(path)

        parts = {'scheme': scheme,
                 'user': user,
                 'password': password,
                 'host':host, 'port':port,
                 'path': path, 'params': params,
                 'fragment':fragment
                 }
        self._update_parts(parts)
        self.parsed = True
        
    def get_parts(self):
        # FIXME: to remove
        if not self.parsed:
            self._parse()
        return self.parts

    def _update_parts(self, hash):
        self.parts = hash
        for k, v in self.parts.iteritems():
            setattr(self, k, v)

    def is_valid_media(self):
        # FIXME: to remove
        location = self.get_location()
        return misc.file_is_movie(location) or misc.file_is_music(location) or \
            location.startswith('dvd://') or location.startswith('http://') or \
            location.startswith('daap://')

    def join(self, path):
        """ same as os.path.join """
        current_path = self.path
        if not current_path.endswith('/'):
            current_path += '/'
        current_path += path
        
        parts = self.get_parts()
        parts['path'] = current_path
        return Uri(parts=parts)
        
    def get_location(self):
        # FIXME: should be __repr__
        user_pass = ''
        if self.user:
            user_pass = "%s:" % self.user
            if self.password:
                user_pass += "%s@" % self.password

        netloc = self.host
        if self.port and self.host.find(':') == -1:
            netloc += ":%s" % self.port

        params = self.params or ''
        if params:
            if not self.safe:
                escaped = ','.join('='.join((k,urllib.quote(v)))
                                   for k,v in params.iteritems())
            elif self.quote:
                escaped = '&'.join('='.join((k,urllib.quote(v)))
                                   for k,v in params.iteritems())
            else:
                escaped = '&'.join('='.join((k,v))
                                   for k,v in params.iteritems())

            params = "?%s" % escaped
            
        fragment = self.fragment
        if fragment:
            fragment = "#%s" % fragment
        location = "%s://%s%s%s%s%s" % (self.scheme, user_pass, netloc,
                                        self.path, params, fragment)
        return location

    __str__ = get_location

    def __repr__(self):
        return "'%s'" % self

if __name__ == '__main__':
    for uri in sys.argv[1:]:
        u = Uri(location=uri)
        print u.get_parts()
        assert str(u) == uri, "%r != %r" % (u, uri)
