# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

import copy
from twisted.internet import threads
from elisa.core import media_uri
from elisa.core.signals.misc import state_changed
from elisa.core.utils.misc import handle_failure
from elisa.core import plugin, interfaces
from elisa.core.utils import misc

class MenuItem(object):
    """
    Manipulate menu item element
    """
    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    def __init__(self, short_name="None", parent=None):
        self._items = []
        self._items_count = 0
        self.__enabled = True
        self.__shown = True
        self.set_parent(parent)
        self.set_short_name(short_name)
        self.set_help_string("")
        self.set_icon_path(None)
        self.set_target_path(None)
        self.set_thumbnailer(None)

        self.callbacks = {'action': None,
                          'loadmore': None,
                          }
        
    def __repr__(self):
        return "<%s instance at 0x%x %r: %s>" % (self.__class__.__name__,
                                                 id(self),
                                                 self.get_short_name(),
                                                 self.get_target_path())
    
    def __iter__(self):
        return iter(self._items)

    def get_enabled(self):
        return self.__enabled

    def set_enabled(self, value):
        send_update = self.__enabled != value
        self.__enabled = value
        if send_update:
            state_changed.emit(self, 'enabled', value)

    _enabled = property(fget=get_enabled, fset=set_enabled)

    def disable(self):
        self._enabled = False

    def enable(self):
        self._enabled = True

    def is_enabled(self):
        return self._enabled

    def get_shown(self):
        return self.__shown

    def set_shown(self, value):
        send_update = self.__shown != value
        self.__shown = value
        if send_update:
            state_changed.emit(self, 'shown', value)

    _shown = property(fget=get_shown, fset=set_shown)

    def hide(self):
        self._shown = False

    def show(self):
        self._shown = True

    def is_visible(self):
        return self._shown

    def set_short_name(self, name, send_event=False):
        "name used on menu if picture is not set"
        self._short_name = name
        if send_event:
            state_changed.emit(self, 'icon-label', name)

    def get_short_name(self):
        "name used on menu if picture is not set"
        if not hasattr(self, '_short_name'):
            self.set_short_name('')
        return self._short_name

    def get_items(self, start, count, notify_callback, *args):
        """dynamic fetch the list of sub MenuTree/MenuItem instances
        start : Zero based start item
        count : number of item returned via callback
        notify_callback : callback with menuitem list in parameter
        data : data passed to the callback
        """
        def got_items(items):
            for i in items:
                i.set_parent(self)
            all_items = self._items + items
            if count == -1:
                selected_items = all_items
            else:
                selected_items = all_items[start:start+count]
            notify_callback(selected_items, *args)

        dfr = threads.deferToThread(self.loadmore, start, count)
        dfr.addCallback(got_items)
        dfr.addErrback(handle_failure)
        return dfr

    def get_data_plugin(self, uri):
        # look for the plugin supporting this type of uri
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)
        return data_access_plugin

    def has_items(self):
        has_items = self._items_count > 0
        path = self.get_target_path()
        try:
            target = media_uri.Uri(location=path)
        except TypeError:
            pass
        else:
            if not has_items:
                data_plugin = self.get_data_plugin(target)
                if data_plugin and path:
                    has_items = data_plugin.has_children(path)
                    
        return has_items

    def set_items_count(self, count):
        self._items_count = count

    def get_items_count(self):
        return self._items_count

    def __contains__(self, item):
        if isinstance(item, basestring):
            r = len([i.get_target_path() for i in self._items
                     if i.get_target_path() == item]) > 0
        else:
            r = item in self._items
        return r

    def insert_item(self, index, item, parent=None):
        if item.get_target_path() and item.get_target_path() in self:
            return
        
        if not parent:
            parent = self

        item.set_parent(parent)
        self._items.insert(index, item)
        self.set_items_count(self.get_items_count() + 1)

    def add_item(self, item, parent=None):
        """ Add a new MenuTree/MenuItem in our children list and link
        the item with self
        """
        if item.get_target_path() and item.get_target_path() in self:
            return
        
        if not parent:
            parent = self
            
        item.set_parent(parent)
            
        self._items.append(item)
        self.set_items_count(self.get_items_count() + 1)
    
    def del_item(self, item):
        """ Remove the item instance from our children list
        """
        if item in self._items:
            index = self._items.index(item)
            self._items[index].set_parent(None)
            del self._items[index]
            self.set_items_count(self.get_items_count() - 1)
            
    def del_item_with_target(self, name):
        item = self.get_item_with_target(name)
        if item:
            self.del_item(item)
        return item is not None

    def get_item_with_target(self, path):
        """look for a MenuTree/MenuItem with given target path in the
        children list. This method is recursive
        """
        found = None
        path = misc.fs_decode(path)
        for item in self._items:
            target_path = item.get_target_path()
            if target_path is None:
                continue
            target_path = misc.fs_decode(target_path)
            if target_path == path:
                found = item
            if not found:
                found = item.get_item_with_target(path)
            if found:
                return found

        return None

##     def __repr__(self):
##         items = self.get_items()
##         name = self.get_short_name()
##         return "%s (%s items)" % (name, len(items))

    def get_tree_from_item(self, parent_item=None, root_level=False):
        """ Create a new MenuTree starting at parent and including all
        the items stored by this item.
        """
        if not parent_item:
            parent_item = self.get_root()

        menu = MenuTree(parent_item.get_short_name())

        def build_sub_tree(items, parent):
            for item in items:
                new_item = item.copy()
                menu.add_item(new_item, parent=parent)
                if not root_level:
                    build_sub_tree(item.get_items(),new_item)

        build_sub_tree(parent_item.get_items(), parent_item.copy())
        return menu



    def pretty_print(self):
        """ Textual representation of the item """
        level = self.get_level()
        representation = "%s - %s" % ("  " * level,
                                      self.get_short_name())
        for item in self._items:
            representation += "\n" + item.pretty_print()
        return representation

    def copy(self):
        dup = copy.deepcopy(self)
        dup.set_parent(None)
        dup._items = []
        return dup

    def set_parent(self, parent):
        """ Link the current instance with another MenuTree, which
        becomes our parent node
        """
        self._parent = parent

    def get_parent(self):
        """ MenuTree parent accessor
        """
        return self._parent

    def get_root(self):
        parent = self.get_parent()
        while parent:
            if not parent.get_parent():
                return parent
            parent = parent.get_parent()
        return None

    def get_level(self):
        """ Fetch the depth at which the instance is located in the
        tree. Return 0 if we are root.
        """
        parent = self.get_parent()
        level = 0
        while parent:
            parent = parent.get_parent()
            level += 1
        return level

    def set_help_string(self, help_string):
        """helpstring shown on box when item is selected"""
        self._help_string = help_string

    def get_help_string(self):
        """helpstring shown on box when item is selected"""
        return self._help_string

    def set_icon_path(self, path, send_event=False):
        """ """
        self._icon_path = path
        if send_event:
            state_changed.emit(self, 'icon-path', path)
            
    def get_icon_path(self):
        """ """
        return self._icon_path

    def set_thumbnailer(self, thumbnailer):
        """ """
        self._thumbnailer = thumbnailer
            
    def get_thumbnailer(self):
        """ """
        return self._thumbnailer

    def set_target_path(self, path):
        """ """
        self._target_path = path

    def get_target_path(self):
        """ """
        return self._target_path

    def set_loadmore_callback(self, callback, **kw):
        self.callbacks['loadmore'] = (callback, kw)
        #print "set callback %s with %s"%(callback, kw)

    def loadmore(self, start, count):
        if not self.callbacks['loadmore']:
            return []

        callback, kw = self.callbacks['loadmore']
        kw['start_item'] = start
        kw['nb_items'] = count
        kw['menu_item'] = self
        return callback(self.get_target_path(), **kw)

    def fire_callback(self, name, *args):
        callback = self.callbacks.get(name)
        if callback and callable(callback):
            args = args or (self,)
            callback(*args)

    def has_callback(self, name):
        return bool(self.callbacks.has_key(name) and self.callbacks[name])

    def get_action_callback(self):
        return self.callbacks.get('action')

    def set_action_callback(self, callback, send_event=False):
        """callback called when menu item is activated"""
        self.callbacks['action'] = callback
        if send_event:
            state_changed.emit(self, 'action-callback', callback)

    def fire_action(self, *args):
        """message called when menu item is activated"""
        self.fire_callback('action', *args)



class MenuTree(MenuItem):
    """
    manipulate a tree
    """

    def get_parent(self):
        return None

    def get_root(self):
        return self

    def copy(self):
        """ Create a copy of the MenuItem
            TODO: copy the items too ?
        """
        dup = copy.deepcopy(self)
        dup._items = []
        return dup

    def as_menu_item(self):
        new_item = MenuItem(short_name=self.get_short_name())
        new_item._items = self._items
        return new_item

    def pretty_print(self):
        """ Textual representation of the tree. This method is recursive
        """
        items = self._items
        representation = "- %s (%s items)" % (self.get_short_name(),
                                              len(items))
        for item in items:
            representation += "\n" + item.pretty_print()
        return representation


if __name__ == '__main__':
    root = MenuTree()

    root.add_item(MenuItem(short_name="level 1"))

    sub = MenuItem(short_name="level 1 bis")
    sub.add_item(MenuItem(short_name="level 2"))
    root.add_item(sub)

    print root.pretty_print()
    print "-" * 80
    lvl1b = root.get_item_with_name("level 1 bis")
    print lvl1b

    print root.get_items()

    print sub.get_root()

    new_sub_tree = root.get_tree_from_item()#lvl1b)
    print new_sub_tree
    print new_sub_tree.get_items()[0].get_level()
    print new_sub_tree.pretty_print()

    print root.get_tree_from_item(lvl1b).pretty_print()

    print "-" * 80

    print root.get_tree_from_item(root_level=True).pretty_print()
