# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import log, media_uri
from elisa.core.menu import MenuItem
from twisted.internet import reactor, defer
from elisa.core.utils.misc import handle_failure
from elisa.core.signals.menu import *
from elisa.core.signals.skin import *
from elisa.core import history_item
import mutex

"""
TODO:

- widget cache

DONE:

- basic navigation
- signal management
- emit active-changed when we go up/down a level
- item activation
- SignalEmitter
- navigation history
- dynamic menu

"""

# FIXME: remove this when progressive menu loading is activated
MENU_ITEMS_NB=100

class MenuManager:

    def __init__(self, root_menu_item):
        self.logger = log.Logger()
        
        self._mutex = mutex.mutex()
        self.lock()
        
        self._root_menu_item = root_menu_item
        self._root_history_menu_item = history_item.HistoryItem(root_menu_item)
        self._history = [ self._root_history_menu_item ]
        new_menu_item.connect(self._new_menu_item)
        del_menu_item.connect(self._del_menu_item)
        
    def _new_menu_item(self, menu_item, index):
        parent = menu_item.get_parent()
        parent_hist = self.get_history_item_for_menu_item(parent)
        if parent_hist and parent_hist.get_nb_children():
            self.add_menu_item(menu_item, parent, index)

    def _del_menu_item(self, menu_item):
        self.remove_menu_item(menu_item)
        
    def lock(self):
        r = self._mutex.testandset()
        if r:
            self.logger.debug("Locking menu_manager")
        return r

    def is_locked(self):
        return self._mutex.locked
    
    def unlock(self):
        self.logger.debug("Unlocking menu_manager")
        self._mutex.unlock()
        #self._history[0].pprint()

    def build_root_menu(self):
        self.logger.debug("Building root menu")
        dfr = self._root_history_menu_item.get_menuitem().get_items(0, -1,
                                                            self._get_items_cb,
                                                            True, -1, None)
        # self.set_current_item(level)
        # level.set_next_child_active()
        dfr.addCallback(self._root_menu_built)
        dfr.addErrback(handle_failure)
        return dfr
        
    def _root_menu_built(self, r):
        root_menu_built.emit(self._root_history_menu_item.get_menuitem())
        self.unlock()
        
    def get_root_item(self):
        return self._root_menu_item

    def _get_current_hist_item(self):
        level = self._history[-1]
        return level

    def _set_current_hist_item(self, hist_item):
        if hist_item not in self._history:
            self._history.append(hist_item)
            
    def get_current_menu_item(self):
        return self._get_current_hist_item().get_menuitem()

    def get_menu_item_children(self, menu_item, start=0, count=-1):
        """
        returns a list of MenuItem instances stored under the given menu_item
        """
        hist_item = self.get_history_item_for_menu_item(menu_item)
        children = [ child.get_menuitem() for child in hist_item.get_children() ]
        if count == -1:
            children = children[start:]
        else:
            children = children[start:count]
        return children

    def activate_current_menu_item(self):
        """
        fire the action callback on the currently selected menu item.
        emits 'item-activated' signal with menu_item and active widget
        in parameters
        """
        current_item = self._get_current_hist_item()
        menu_item = current_item.get_menuitem()
        widget = current_item.get_widget()
        has_action = menu_item.has_callback('action')
        if has_action:
            item_activated.emit(menu_item, widget)
            menu_item.fire_action()
        return has_action
    
    def go_up_level(self, record=True):
        """
        navigate to up level and record path to history
        """
        if not self.lock():
            return False
        current_item = self._get_current_hist_item()
        #old_widget = current_item.get_active_child().get_widget()
        #old_menu_item = current_item.get_active_child().get_menuitem()
        old_widget = current_item.get_widget()
        old_menu_item = current_item.get_menuitem()
        parent_item = current_item.get_parent()
        if parent_item:
            self.logger.debug("Going up level from %r" % current_item)
            widget = parent_item.get_widget()
            menu_item = parent_item.get_menuitem()
            direction = u'vertical'
            step = -1

            if len(self._history) > 1:
                level = self._history[-1]
                self._remove_hist_level(level)

            self._set_current_hist_item(parent_item)
            active_changed.emit(old_menu_item, old_widget,
                                menu_item, widget, direction, step)
        return True
    
    def go_down_level(self, recorded=True, lock=True, requested_menu_items=-1):
        """
        navigate to the down level and optionally go to the last
        recorded item
        """
        if lock and not self.lock():
            return None
        current_item = self._get_current_hist_item()
        menu_item = current_item.get_menuitem()
        
        dfr = None
        if current_item.get_parsed():
            self.logger.debug("Going again down from %r" % menu_item)
            dfr = menu_item.get_items(0, MENU_ITEMS_NB, self._get_items_cb,
                                      False, requested_menu_items, menu_item)
        elif menu_item: # and menu_item.has_items():
            self.logger.debug("Going down level from %r" % menu_item)
            dfr = menu_item.get_items(0, MENU_ITEMS_NB, self._get_items_cb,
                                      False, requested_menu_items, menu_item)
        if not dfr:
            self.logger.debug("Can't go down from %r at %r" % (menu_item,
                                                               current_item, menu_item))
        else:
            level_loading.emit(menu_item)

        return dfr

    def go_next_menu_item(self, do_lock=True):
        """
        increment the index of the active menu item
        """
        if do_lock and not self.lock():
            return False        
        parent_item = self._get_current_hist_item().get_parent()
        old_widget = None
        old_menu_item = None
        if parent_item:
            old_widget = parent_item.get_active_child().get_widget()
            old_menu_item = parent_item.get_active_child().get_menuitem()
        else:
            parent_item = self._root_history_menu_item
            
        item = parent_item.set_next_child_active()
        if item:
            self.logger.debug("Going next item from %r" % old_menu_item)
            menu_item = item.get_menuitem()
            widget = item.get_widget()
            direction = u'horizontal'
            step = 1
            if len(self._history) > 1:
                self._history[-1] = item
            else:
                self._history.append(item)
            active_changed.emit(old_menu_item, old_widget,
                                menu_item, widget, direction, step)
        return True
    
    def go_previous_menu_item(self, do_lock=True):
        """
        decrement the index of the active menu item
        """
        if do_lock and not self.lock():
            return False        
        parent_item = self._get_current_hist_item().get_parent()
        old_widget = parent_item.get_active_child().get_widget()
        old_menu_item = parent_item.get_active_child().get_menuitem()
        item = parent_item.set_previous_child_active()
        if item:
            self.logger.debug("Going previous item from %r" % old_menu_item)
            menu_item = item.get_menuitem()
            widget = item.get_widget()
            direction = u'horizontal'
            step = -1
            if len(self._history) > 1:
                self._history[-1] = item
            else:
                self._history.append(item)
            active_changed.emit(old_menu_item, old_widget,
                                menu_item, widget, direction, step)
        return True
    
    def go_menu_item_with_index(self, index):
        parent_item = self._get_current_hist_item().get_parent()
        old_widget = None
        old_menu_item = None
        if parent_item:
            old_widget = parent_item.get_active_child().get_widget()
            old_menu_item = parent_item.get_active_child().get_menuitem()
        else:
            parent_item = self._root_history_menu_item

        if index in range(parent_item.get_nb_children()):
            parent_item.set_active_index(index)
            hist_item = parent_item.get_active_child()
            self.logger.debug("Going to index %s" % index)
            menu_item = hist_item.get_menuitem()
            widget = hist_item.get_widget()
            direction = u'horizontal'
            step = index
            if len(self._history) > 1:
                self._history[-1] = hist_item
            else:
                self._history.append(hist_item)
            active_changed.emit(old_menu_item, old_widget,
                                menu_item, widget, direction, step)

    def get_menu_item_from_uri(self, uri):
        
        for hist_item in self._history:
            menu_item = hist_item.get_menuitem()
            try:
                item_uri = str(media_uri.Uri(location=menu_item.get_target_path()))
            except TypeError:
                continue
            if item_uri == uri:
                return menu_item
        return None

    def get_menu_item_from_index(self, index, menu_level = None):
        if menu_level:
            parent_hist_item = self.get_history_item_for_menu_item(menu_level)
        else:
            parent_hist_item = self._get_current_hist_item().get_parent()
        menu_item = None
        if index >= 0:
            try:
                hist_item = parent_hist_item.get_children()[index]
            except IndexError:
                pass
            else:
                menu_item = hist_item.get_menuitem()            
        return menu_item
    
    def get_index_from_menu_item(self, menu_item):
        index = -1
        hist_item = self.get_history_item_for_menu_item(menu_item)
        if hist_item:
            hist_item = hist_item.get_parent()
            index = 0
            for child in hist_item.get_children():
                if child.get_menuitem() == menu_item:
                    return index
                index += 1
        return index

    def get_menu_item_for_widget(self, widget):
        for hist_item in self._history:
            w = hist_item.get_widget()
            if w == widget:
                return hist_item.get_menuitem()
        return None

    def add_menu_item(self, menu_item, parent_menu_item, index=None):
        self.logger.debug("Adding menu item %r in %r" % (menu_item,
                                                         parent_menu_item))
        hist_item = self.get_history_item_for_menu_item(parent_menu_item)
        return self._add_hist_item(menu_item, hist_item, index=index)

    def remove_menu_item(self, menu_item):
        parent = menu_item.get_parent()
        hist_item = self.get_history_item_for_menu_item(parent)
        self._remove_hist_item(menu_item, hist_item)

    def _add_hist_item(self, menu_item, level, index=None, emit_signal=True):
        """
        a new menu item under parent's level and emit 'item-added'
        signal
        """
        item = None
        if menu_item != level.get_menuitem():
            item = history_item.HistoryItem(menu_item)
            level.add_child(item, index)
            self.logger.debug("Added new item %r in level %r" % (item, level))
            if index is None:
                index = -1
            if emit_signal:
                item_added.emit(item.get_menuitem(), index)
        else:
            self.logger.debug("Could not add %r to %r" % (menu_item, level))
        return item
    
    def _remove_hist_item(self, menu_item, level):
        """
        emit 'item-removed' signal, remove the widget, associated menu
        item from its level
        """
        removed = False
        index = 0
        active_index = level.get_active_index()
        for child in level.get_children():
            if child.get_menuitem() == menu_item:
                if index < active_index:
                    level.set_active_index(active_index-1)
                self.logger.debug("Removed item %r from %r" % (child, level))
                item_removed.emit(child.get_menuitem())
                if child.has_level():
                    self.remove_menu_level(menu_item)
                level.del_child(child)
                removed = True
                break
            index += 1
        if not removed:
            self.logger.debug("Could not remove %r from %r" % (menu_item, level))
            
    def add_menu_level(self, menu_item):
        hist_item = self.get_history_item_for_menu_item(menu_item)
        self._add_hist_level(hist_item)
    
    def remove_menu_level(self, menu_item):
        hist_item = self.get_history_item_for_menu_item(menu_item)
        self._remove_hist_level(hist_item)

    def _add_hist_level(self, hist_item):
        """
        add new menu level for the given menu item and emit
        'level-added' signal
        """
        new_item = hist_item.get_active_child()
        level_added.emit(new_item.get_menuitem() )
        return new_item

    def _remove_hist_level(self, level):
        """
        emit 'level-removed' signal and remove the widget associated
        to the given menu level
        """
        self.logger.debug("Removing level %s" % level)
        if level in self._history:
            self._history.remove(level)
        level_removed.emit(level.get_menuitem())

    def bind_menu_item_to_widget(self, menu_item, widget):
        """
        bind a menu item with an actual widget
        """
        hist_item = self.get_history_item_for_menu_item(menu_item)
        hist_item.set_widget(widget)
        current_item = self._get_current_hist_item()

        if current_item.get_active_child() == hist_item:
            self.logger.debug("Binding menu_item %r to widget %r" % (menu_item,
                                                                     widget))
            self._set_current_hist_item(hist_item)
##         else:
##             self.logger.debug("Could not bind %r to %r" % (menu_item, widget))
            
        item_bound.emit(hist_item.get_menuitem(), hist_item.get_widget())
        
    def bind_menu_level_to_widget(self, menu_item, widget):
        """
        bind a level residing under the menu item to an actual widget
        """
        hist_item = self.get_history_item_for_menu_item(menu_item)
        if hist_item:
            self.logger.debug("Binding menu_level %r to widget %r" % (menu_item,
                                                                      widget))
            hist_item.set_level_widget(widget)
            self._set_current_hist_item(hist_item)
            level_bound.emit(hist_item.get_menuitem(),
                             hist_item.get_level_widget())
        
    def get_widget_cache_size(self):
        """
        retrieve current cache size as integer
        """
        pass

    def set_widget_cache_size(self, new_size):
        """
        update cache size
        """
        pass

    def expire_cache(self):
        """
        force the cache to purge itself
        """
        pass

    def get_level_widget_for_menu_item(self, menu_item):
        """
        return the level_widget bound to a menu_item.
        if not found, returns None
        """
        current_item = self._history[-1]
        level_widget = None
        hist_item = self.get_history_item_for_menu_item(menu_item)
        if hist_item:
            level_widget = hist_item.get_level_widget()
        #self.logger.debug("Found level widget %r for %r" % (level_widget,
        #                                                    menu_item))
        return level_widget

    def get_widget_item_for_menu_item(self, menu_item):
        """
        return the item_widget bound to a menu_item.
        if not found, returns None
        """
        current_item = self._history[-1]
        item_widget = None
        hist_item = self.get_history_item_for_menu_item(menu_item)
        if hist_item:
            item_widget = hist_item.get_widget()
        #self.logger.debug("Found widget %r for %r" % (item_widget, menu_item))
        return item_widget

    def get_history_item_for_menu_item(self, menu_item):
        current_item = self._history[-1]
        hist_item = None
        while current_item:
            hist_item = current_item.find(menu_item)
            if hist_item:
                break
            current_item = current_item.get_parent()
        if not hist_item:
            # try the root item at last resort
            if self._root_history_menu_item == menu_item:
                hist_item = self._history[0]
        #self.logger.debug("Returning %r" % hist_item)
        return hist_item
                        
    def _get_items_cb(self, menu_items, is_root_menu, requested_menu_items, parent_menu_item):
        """
        history_item.HistoryItem creation callback
        """
        if not menu_items:
            root = self.get_root_item()
            root_hist_item = self.get_history_item_for_menu_item(root)
            plugin_tree = root_hist_item.get_active_child().get_menuitem()
            short_name = {'music': _('No music files found in this directory'),
                          'movies': _('No movies found in this directory'),
                          'pictures': _('No pictures found in this directory'),
                          }.get(plugin_tree.get_name())
                          
            empty_menu_item = MenuItem(short_name,parent_menu_item)
            empty_menu_item.set_icon_path('empty.png')
            menu_items.append(empty_menu_item)
        
        current_hist_item = self._get_current_hist_item()
        first = None
        
        if current_hist_item.get_level_widget():
            level_added.emit(current_hist_item.get_menuitem())
            level = current_hist_item
            first = current_hist_item.get_active_child()
        else:
            added_items = 0
            level = self._add_hist_level(current_hist_item)
            for menu_item in menu_items:
                emit_signal = True
                if requested_menu_items > 0 and added_items > requested_menu_items:
                    emit_signal = False
                hist_item = self._add_hist_item(menu_item, level,
                                                emit_signal=emit_signal)
                if not first:
                    first = hist_item
                added_items += 1
            
        if not is_root_menu:                
            current_hist_item = self._get_current_hist_item()
            parent_item = current_hist_item.get_parent()
            if not parent_item:
                parent_item = current_hist_item
            
            if parent_item:
                old_active = parent_item.get_active_child()
                old_widget = old_active.get_widget()
                old_menu_item = old_active.get_menuitem()

                if not current_hist_item.get_parsed():
                    menu_item = first.get_menuitem()
                    widget = first.get_widget()
                else:
                    menu_item = current_hist_item.get_active_child().get_menuitem()
                    widget = current_hist_item.get_active_child().get_widget()
                    old_active = current_hist_item.get_parent().get_active_child()
                    old_widget = old_active.get_widget()
                    old_menu_item = old_active.get_menuitem()
                    
                direction = u'vertical'
                step = 1
                active_changed.emit(old_menu_item, old_widget,
                                    menu_item, widget, direction, step)

            if not current_hist_item.get_parent().get_parsed():
                current_hist_item.get_parent().set_parsed(True)
            self._set_current_hist_item(current_hist_item.get_active_child())
        level_loaded.emit(level.get_menuitem(), len(menu_items))
