# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

import pygst
pygst.require('0.10')
import gst
import gobject, sys, os
from mutex import mutex

from elisa.core.player import playable, default_queue
from elisa.core import media_uri
from elisa.core import log
from elisa.core import common, plugin, interfaces
from pgm import render
from pgm.form import stream

from elisa.core.utils import misc
from elisa.core.utils.misc import handle_failure
from elisa.core.signals.player import *
from elisa.core.signals.misc import state_changed

from twisted.spread import pb
from twisted.internet import task , defer

from pgm import log as pgm_log
#pgm_log.set_level(pgm_log.DEBUG)


class Player(pb.Referenceable):

    PLAYING = gst.STATE_PLAYING
    PAUSED = gst.STATE_PAUSED
    STOPPED = gst.STATE_READY

    config = {'seek_forward':'c',
              'seek_backward': 'x',
              'increment_volume': 'u',
              'decrement_volume': 'i',
              'increment_playback_speed': 'j', 'toggle_play_pause': 'p',
              'decrement_playback_speed': 'k', 'pause': 'l', 'stop': 's',
              'next': 'v', 'previous': 'w', 'toggle_mute': 'm',
              }

    config_extra_params = {'increment_volume': (3,),
                           'decrement_volume': (3,),
                           'increment_playback_speed': (0.1,),
                           'decrement_playback_speed': (0.1,)
                           }

    default_config = {'subtitle_font': 'Sans Bold 24',
                      'show_visualization': '0',
                      'seek_forward_step': '60',
                      'seek_backward_step': '10',
                      'audio_sink_name': 'autoaudiosink',
                      'visualization_name': 'libvisual_plazma'}
    
    data_access = plugin.ExtensionPoint(interfaces.IDataAccess)

    def __init__(self, uri='', queue=None):
        self.logger = log.Logger()
        self._update_stats_loop = None
        self._stream = None
        
        win = common.get_application().get_window()
        if win:
            win.sig_key_press_event.connect(self.key_pressed)
                
        self._playbin = None
        self._playback_speed = 1.0
        
        self.playing_uri = None
        self.head_uri = None
        self.current_playable = None

        self.set_queue(queue)
        self.set_uri(uri)

        self.set_key_actions()
        
    def set_key_actions(self):
        appli = common.get_application()
        config = appli.get_config()
        pgmrender = render.get_renderer()

        self.key_actions = {}
        for opt_name, default_value in self.config.iteritems():
         
            value = config.get_option('%s_key' % opt_name, section='player',
                                      default=default_value)
            params = self.config_extra_params.get(opt_name, ())
            try:
                key_value = getattr(pgmrender, 'KEY_%s' % value)
            except AttributeError:
                self.logger.info("Skipping configuration of %r keybinding" % value)
                continue
            self.key_actions[key_value] = (opt_name, params)
            config.set_option('%s_key' % opt_name, value, section='player')

        for opt_name, default_value in self.default_config.iteritems():
            option = config.get_option(opt_name, section='player')
            if not option:
                config.set_option(opt_name, default_value, section='player')

        seek_forward_step = int(config.get_option('seek_forward_step',
                                                  section='player',
                                                  default='60'))
        seek_backward_step = int(config.get_option('seek_backward_step',
                                                   section='player',
                                                   default='10'))
        self.fs_key_actions = {
            pgmrender.KEY_UP : ('increment_volume',(3,)),
            pgmrender.KEY_DOWN: ('decrement_volume', (3,)),
            pgmrender.KEY_RIGHT: ('seek_forward', (seek_forward_step,)),
            pgmrender.KEY_LEFT: ('seek_backward', (seek_backward_step,))
            }


    def get_data_plugin(self, location):
        # extract the type or uri (file:// .. smb:// ....)
        if not isinstance(location, media_uri.Uri):
            uri = media_uri.Uri(location=location)
        else:
            uri = location

        # look for the plugin supporting this type of uri
        data_access_plugin = self.data_access.get("uri:%s" % uri.scheme)

        return data_access_plugin
        
    def get_queue(self):
        return self.queue

    def set_queue(self, queue):
        if not queue:
            queue = default_queue.DefaultQueue()
        self.queue = queue

    def get_id(self):
        if self._uri:
            return id(self._uri)
        return id(self)

    def update_status(self):
        """ Update current playing item status, lenght and volume
        """
        position, duration = self.get_status()
        volume = self.get_volume()
        
        # update playing item informations
        current_item = self.get_current_item()
        if current_item:
            current_item.set_length(duration)
            current_item.set_status(position)
            current_item.set_volume(volume)

            uri = media_uri.Uri(current_item.get_uri())

            duration = float(duration)
            if duration > 0.:
                fraction = position/float(duration)
            else:
                fraction = 0.
            
            stats = (uri.get_label(), current_item.get_status_string(),
                     fraction)
            state_changed.emit(current_item, 'player-stats', stats)
            
        # poll the gst bus for messages
        bus = self._playbin.get_bus()
        message = bus.poll(gst.MESSAGE_ERROR|gst.MESSAGE_EOS|gst.MESSAGE_TAG,
                           timeout=1)
        if message:
            self.on_message(bus, message)
        

    def on_message(self, bus, message, extra=None):
        """ Callback used by Gstreamer messaging system
        """
        t = message.type
        
        if t == gst.MESSAGE_TAG:
            audio_sink = None
            video_sink = None
            try:
                audio_sink = self._playbin.get_property('audio-sink')
            except: pass
            try:
                video_sink = self._playbin.get_property('video-sink')
            except: pass
            parsed_tag = dict(message.parse_tag())
            audio_ok, video_ok = True, True
            audio_state, video_state = self.PLAYING, self.PLAYING
            
            if audio_sink:
                f, audio_state, f = audio_sink.get_state()
            if video_sink:
                f, video_state, f = video_sink.get_state()
            

            if 'audio-codec' in parsed_tag and audio_state != self.PLAYING:
                audio_ok = False
            if 'video-codec' in parsed_tag and video_state != self.PLAYING:
                video_ok = False

            src = self._playbin
            error_msg = ''
            if not audio_ok:
                error_msg = _('No valid audio codec found')
                src = audio_sink
            elif not video_ok:
                error_msg = _('No valid video codec found')
                src = video_sink
                
            if error_msg:
                struct = gst.Structure("custom")
                struct['error'] = error_msg
                struct['code'] = gst.STREAM_ERROR_CODEC_NOT_FOUND
                msg = gst.message_new_custom(gst.MESSAGE_ERROR, src, struct)
                bus.post(msg)
            
        if t == gst.MESSAGE_ERROR:
            err = None
            debug_msg = ""
            if message.src != self._playbin:
                class M: pass
                err = M()
                try:
                    err.message = message.structure['error']
                except:
                    err.message = ""
                try:
                    err.code = message.structure['code']
                except:
                    err.code = 0
            else:
                err, debug_msg = message.parse_error()
            if err:
                print "Error: %s %s" % (err.message, debug_msg)
                error_signal.emit(err)
                self.on_eos()
        elif t == gst.MESSAGE_EOS:
            self.on_eos()
            
    def key_pressed(self, key_event):
        val = key_event.keyval
        action, params = None, ()
        skin = common.get_application().get_skin()
        pgmrender = render.get_renderer()
        
        #check if more than one other task is already running
        if len(common.get_application().get_task_manager()) > 1:
            self.logger.info("Not catching key press because another task is running")
            return
        
        if self.get_playing_uri():
            if val in self.key_actions:
                action, params = self.key_actions[val]
            elif val in (pgmrender.KEY_UP, pgmrender.KEY_DOWN) \
                     and skin.dock_visible():
                action, params = self.fs_key_actions[val]
            elif val in self.fs_key_actions \
                     and not skin.menu_visible():
                if not skin._dock._control_menu or not skin._dock._control_menu.visible:
                    action, params = self.fs_key_actions[val]
        if action:
            getattr(self, action)(*params)

    def on_eos(self):
        if not self.next():
            self.stop()

    def close(self):
        self.logger.info("Closing player")
        if self._playbin:
            self._playbin.set_state(gst.STATE_NULL)

    def get_playing_uri(self):
        return self.playing_uri

    def get_stream(self):
        return self._stream

    def get_config(self):
        config = common.get_application().get_config()
        return config.get_section('player')
        
    def set_playing_uri(self, uri, short_name=None):
        """ Play the media identified by its uri string
        """
        dfr = defer.Deferred()
        self.playing_uri = uri
        self.old_fraction = 0.
        
        if uri:
            uri = misc.fs_encode(str(uri))
            
            if self.get_state() != self.PAUSED and self._playbin:
                self._playbin.set_state(gst.STATE_NULL)
            self.add_uri(uri, short_name)

            def launch(typed_stream):
                self._playbin = typed_stream.pipeline
                discoverer = typed_stream.discoverer
                has_audio = discoverer.is_audio
                has_video = discoverer.is_video
                #has_video = has_audio = True
                extra = {'contains_audio':has_audio, 'contains_video':has_video}
                self.current_playable = playable.Playable(uri, **extra)

                # try to load subtitle
                if has_video:
                    subtitle_uri = self.current_playable.get_subtitle_uri()
                    if subtitle_uri:
                        self.logger.info("Loading subtitle: %s" % subtitle_uri)
                        self._playbin.set_property('suburi', subtitle_uri)
                        sub_font = self.get_config().get('subtitle_font',
                                                         'Sans Bold 24')
                        self._playbin.set_property('subtitle-font-desc',
                                                   sub_font)

                if has_audio:
                    try:
                        asink = self._playbin.get_property('audio-sink')
                    except TypeError:
                        # FIXME: there's no .has_property()????
                        pass
                    else:
                        audio_sink_name = self.get_config().get('audio_sink_name',
                                                                'autoaudiosink')
                        audio_sink = gst.element_factory_make(audio_sink_name,
                                                              'audio-sink')
                        self._playbin.set_property('audio-sink', audio_sink)
                                                            

                self._stop_sending_updates()
                self._start_sending_updates()
                self.un_mute()
                try:
                    dfr.callback(True)
                except defer.AlreadyCalledError:
                    pass

            gst_pipeline = None
            
            data_access = self.get_data_plugin(uri)
            if data_access:
                gst_pipeline = data_access.get_gst_pipeline()
                

            self._stream = stream.Stream(pipeline=gst_pipeline,
                                         location=uri,
                                         discovered_callback=launch)
            #launch(self._stream)
        else:
            self._stop_sending_updates()
        return dfr
    
    def set_head_uri(self, uri):
        self.head_uri = str(uri)

    def get_head_uri(self):
        return self.head_uri

    def set_uri(self, uri, short_name=None):
        dfr = None
        if uri:
            self.stop()
            uri = media_uri.Uri(location=uri)

            data_plugin = self.get_data_plugin(uri)
            if data_plugin and data_plugin.is_dir(str(uri)):
                self.set_head_uri(uri)
            else:
                dfr = self.set_playing_uri(uri, short_name)
        return dfr
        
    remote_set_uri = set_uri
    
    def play_uri_once(self, uri, short_name=None):
        if self.get_state() in (self.PLAYING, self.PAUSED):
            stopped_signal.emit(self.get_current_item())

        if not short_name:
            short_name = media_uri.Uri(location=uri).get_label()
            
        def launch(r):
            self.play(start=True)
            self.logger.info("Playing URI: %s" % uri)
            
            
        dfr = self.set_uri(uri, short_name)
        if dfr:
            dfr.addCallback(launch)
            dfr.addErrback(handle_failure)

    remote_play_uri = play_uri_once

    def play(self, send_stats=True, start=False):
        """ Set the player state to playing mode and fire a Playing
        event holding the current playing item of the queue.
        """
        if not self._playbin:
            return
        self.logger.info("Playing")
        self._playbin.set_state(gst.STATE_PLAYING)

        item = self.get_current_item()
        if item:            
            if send_stats:
                self._start_sending_updates()
            else:
                self._stop_sending_updates()
                state_changed.emit(item, 'player-stats', '')
        if self.get_playback_speed() != 1.:
            self.set_playback_speed(1.0)
        if item:
            playing_signal.emit(item, start)

    remote_play = play

    def pause(self):
        """ Set the player state to paused mode and fire a Paused
        event holding the current playing item of the queue.
        """
        if not self._playbin:
            return
        self.logger.info("Pausing")
        self._playbin.set_state(gst.STATE_PAUSED)
        item = self.get_current_item()
        paused_signal.emit(item)
        self._stop_sending_updates(True)

    remote_pause = pause

    def toggle_play_pause(self):
        if self.get_state() == self.PLAYING:
            self.pause()
        else:
            self.play()

    remote_toggle_play_pause = toggle_play_pause
                        
    def _start_sending_updates(self):
        if not self._update_stats_loop:
            self._update_stats_loop = task.LoopingCall(self.update_status)
        if not self._update_stats_loop.running:
            self._update_stats_loop.start(1.0)

    def _stop_sending_updates(self, pause=False):
        if self._update_stats_loop and self._update_stats_loop.running:
            self._update_stats_loop.stop()
            if not pause:
                self._update_stats_loop = None
                
    def stop(self):
        """ Set the player state to stopped mode and fire a Stopped
        event.
        """
        if not self._playbin:
            return
        self.logger.info("Stopping")
        self._playbin.set_state(gst.STATE_READY)
        self._playbin = None

        last_item = self.get_current_item()
        if last_item:
            stopped_signal.emit(last_item)
            self.get_queue().set_current_item(None, None)
            self.set_playing_uri(None)
            state_changed.emit(last_item, 'player-stats', '')

    remote_stop = stop

    def seek_forward(self, seconds_or_fraction=None):
        """ Seek the current playing item some seconds
        forward. Seconds are supplied as integers. If the resulting
        location is valid within the playing item's length, the seek is
        performed. Otherwise, nothing happens.
        """
        playable = self.get_current_item()
        seconds = None
        fraction = None
        if seconds_or_fraction is None:
            # use default config
            appli = common.get_application()
            config = appli.get_config()
            seconds = int(config.get_option('seek_forward_step',
                                            section='player', default='60'))
        elif type(seconds_or_fraction) == int:
            seconds = seconds_or_fraction
        else:
            fraction = seconds_or_fraction
            fraction = self.old_fraction + fraction
            if fraction >= 1.:
                fraction = 0.99
        if playable:
            new_location = 0
            if seconds:
                self.logger.info("Seeking %s seconds forward" % seconds)
                new_location = playable.get_status() + (gst.SECOND * seconds)
            elif fraction != None:
                self.logger.info("Seeking to %s location" % fraction)
                new_location = playable.get_length() * fraction
                self.old_fraction = fraction
            if new_location <= playable.get_length() and new_location > 0:
                self.seek_to_location(new_location)

    remote_seek_forward = seek_forward

    def seek_backward(self, seconds_or_fraction=None):
        """ Seek the current playing item some seconds
        backward. Seconds are supplied as integers. If the resulting
        location is valid, the seek is performed. Otherwise, nothing
        happens.
        """
        playable = self.get_current_item()
        seconds = None
        fraction = None
        if seconds_or_fraction is None:
            appli = common.get_application()
            config = appli.get_config()
            seconds = int(config.get_option('seek_backward_step',
                                            section='player', default='10'))
        elif type(seconds_or_fraction) == int:
            seconds = seconds_or_fraction
        else:
            fraction = seconds_or_fraction
            fraction = self.old_fraction - fraction
            if fraction < 0.:
                fraction = 0.
        if playable:
            new_location = 0
            if seconds:
                self.logger.info("Seeking %s seconds backward" % seconds)
                new_location = playable.get_status() - (gst.SECOND * seconds)
            elif fraction != None:
                self.logger.info("Seeking to %s location" % fraction)
                new_location = playable.get_length() * fraction
                self.old_fraction = fraction
            if new_location >= 0:
                self.seek_to_location(new_location)

    remote_seek_backward = seek_backward

    def set_playback_speed(self, speed):
        speed = float(speed)
        if speed > -0.1 and speed <= 0.:
            if self.get_playback_speed() == -0.1:
                speed = 0.1
            else:
                speed = -0.1
        if speed >= 0. and speed < 0.1:
            speed = 0.1
        self.logger.info("Setting playback speed to %s" % speed)
        playable = self.get_current_item()
        if playable:
            location = playable.get_status()
            self._playback_speed = speed
            settings_signal.emit('playback-speed', speed)
            self.seek_to_location(location, speed)

    def get_playback_speed(self):
        return self._playback_speed

    def set_playback_speed_incr(self, delta):
        new_speed = float(self.get_playback_speed() + delta)
        if new_speed > -10.0 and new_speed < 10.0:
            self.set_playback_speed(new_speed)

    def increment_playback_speed(self, step=1):
        self.set_playback_speed_incr(step)

    remote_increment_playback_speed = increment_playback_speed

    def decrement_playback_speed(self, step=1):
        self.set_playback_speed_incr(-step)

    remote_decrement_playback_speed = decrement_playback_speed
        
    def seek_to_location(self, location, speed=None):
        """ Seek to an absolute location in the current playing item.
        @param location: time to seek to, in seconds
        """
        if speed == None:
            speed = self._playback_speed
        if location == gst.CLOCK_TIME_NONE or not self._playbin:
            return
        human_loc = gst.TIME_ARGS(location)
        self.logger.info("Seeking to location : %s at speed %s" % (human_loc,
                                                                   speed))

        if speed > 0.:
            start = location
            end_type = gst.SEEK_TYPE_NONE
            end = 0
        else:
            start = 0
            end = location
            end_type = gst.SEEK_TYPE_SET
            
        try:
            event = self._playbin.seek(speed, gst.FORMAT_TIME,
                                       gst.SEEK_FLAG_FLUSH,
                                       gst.SEEK_TYPE_SET, start,
                                       end_type, end)
        except OverflowError:
            pass
        else:
            if event:
                playable = self.get_current_item()
                playable.set_status(location)
                self._playbin.set_new_stream_time(0L)
                timeout = 40 * gst.MSECOND
                self._playbin.get_state(timeout=timeout)
      
    def increment_volume(self, step=3):
        self.set_volume_incr(step)
        
    def decrement_volume(self, step=3):
        self.set_volume_incr(-step)
        
    def set_volume(self, volume):
        """ Set the volume level of the gst playbin. between 0 and 100%
        """
        if int(volume) in range(101) and self._stream:
            volume = int(volume)
            self.logger.info("Setting volume to %s" % volume)
            item = self.get_current_item()
            if item is not None:
                item.set_volume(volume)
                settings_signal.emit('volume', volume)
                volume = (volume * 10) / 100.
                if volume < 0.:
                    volume = 0.
                elif volume > 10.:
                    volume = 10.
                self._stream.volume = volume
                
    remote_set_volume = set_volume
    
    def get_volume(self):
        volume = 0
        if self._stream:
            volume = self._stream.volume
            volume = (volume * 100) / 10.
        return int(volume)
        
    remote_get_volume = get_volume
    
    def get_mute(self):
        """ return Mute state of the player
        """
        if self._stream:
            return self._stream.mute
        return False

    remote_get_mute = get_mute
               
    def mute(self):
        """ Mute the sound of the player
        """
        if self._stream:
            self.logger.info("Muting")
            self._stream.mute = True
            settings_signal.emit('volume', 0)

    remote_mute = mute
      
    def un_mute(self):
        """ Un-Mute the sound of the player
        """
        if self._stream:
            self.logger.info("UnMuting")
            self._stream.mute = False
            settings_signal.emit('volume', self.get_volume())

    remote_un_mute = un_mute
    
    def toggle_mute(self):
        if self._stream:
            muted = self._stream.mute
            if muted:
                self.un_mute()
            else:
                self.mute()

    def set_volume_incr(self, value):
        self.set_volume(self.get_volume() + value)

    def add_uri(self, uri, short_name=None):
        """ Add a new item in the end of the playing queue.
        """
        self.logger.info("Adding new uri: %s" % uri)
        self.get_queue().set_current_item(uri, short_name)

    def next(self, at_user_request=False):
        """ Move forward to the next item in the playing queue. If the
        end of queue is reached, an IndexError is raised.
        """
        self.logger.info("Next item in queue")

        current = self.get_current_item()
        if current:
            stopped_signal.emit(current)
        
        queue = self.get_queue()
        next_uri, short_name = queue.next_item(self.get_head_uri(),
                                               at_user_request=at_user_request)
        if next_uri:
            queue.set_current_item(next_uri, short_name)
            self.play_uri_once(next_uri, short_name)
            return True
        return False

    remote_next = next

    def previous(self):
        """ Move backward to the previous item in the playing
        queue. If the beginning of the queue is reached an IndexError
        is raised.
        """
        self.logger.info("Previous item in queue")
        prev_uri, short_name = self.get_queue().previous_item()
        if prev_uri:
            self.get_queue().set_current_item((prev_uri, short_name))
            self.play_uri_once(prev_uri, short_name)
            return True
        return False

    def get_current_item(self):
        """ Return the item of the playing queue which is currently selected
        """
        queue = self.get_queue()
        item = self.current_playable
        if queue:
            current_uri, short_name = queue.get_current_item()
            if current_uri:
                try:
                    if not self.current_playable or \
                           (current_uri != self.current_playable.get_uri()):
                        item = self.current_playable = playable.Playable(current_uri)
                except:
                    item = self.current_playable = playable.Playable(current_uri)
        return item

    remote_get_current_item = get_current_item

    def get_status(self):
        "Returns a (position, duration) tuple"
        try:
            position, format = self._playbin.query_position(gst.FORMAT_TIME)
        except:
            position = gst.CLOCK_TIME_NONE

        try:
            duration, format = self._playbin.query_duration(gst.FORMAT_TIME)
        except:
            duration = gst.CLOCK_TIME_NONE

        return (position, duration)

    remote_get_status = get_status

    def get_status_informations(self):
        infos = "State   : %s\n" % self.get_state()
        current_item = self.get_current_item()
        if current_item:
            infos += current_item.get_status_informations()
        return infos

    remote_get_status_informations = get_status_informations

    def get_state(self):
        """ Fetch the current state of the player, can be one of:

        - Player.PLAYING
        - Player.PAUSED
        - Player.STOPPED
        """
        if not self._playbin:
            state = Player.STOPPED
        else:
            state = self._playbin.get_state()[1]
        return state
        
    def get_readable_state(self):
        """ Fetch the current state of the player, can be one of:

        - Player.PLAYING
        - Player.PAUSED
        - Player.STOPPED
        
        and return it as text PLAYING, PAUSED or STOPPED
        """
        if self._playbin:
            state = self._playbin.get_state()[1]
            if state == Player.PLAYING:
                return 'PLAYING'
            elif state == Player.PAUSED:
                return 'PAUSED'

        return 'STOPPED'
        
    remote_get_readable_state = get_readable_state

    def fullscreen(self):
        " should be implemented by SurfacePlayer "

    def set_visualization_plugin(self, stop=False):
        if not self.is_visualization_enabled():
            return False
        self.logger.info("Setting visualization plugin, stopping: %s" % stop)
        if stop and self._playbin:
            self._playbin.set_property('vis-plugin', None)
            return
        r = gst.registry_get_default()
        l = [x for x in r.get_feature_list(gst.ElementFactory) \
             if (gst.ElementFactory.get_klass(x) == "Visualization")]
        option = self.get_config().get('visualization_name', 'goom')
        e = [y for y in l if (y.get_name() == option)]
        if e:
            visplug = gst.element_factory_make(e.pop().get_name())
            if self._playbin:
                self._playbin.set_property('vis-plugin', visplug)
                return True
        return False

    def is_visualization_enabled(self):
        option = int(self.get_config().get('show_visualization', '1'))
        return option

if __name__ == '__main__':

    # usage:
    # player.py uri1 uri2

    mainloop = gobject.MainLoop()


    """
    # one player per uri
    for uri in sys.argv[1:]:

        p = Player(uri)
        p.register('player.playing', on_play)
        manager.add_player(p)
        p.play()
    """

    #"""
    # one player for all uris => use the queue
    p = Player()

    for uri in sys.argv[1:]:
        p.add_uri(uri)

    p.next()
    #"""

    mainloop.run()
