# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core.player.default_queue import DefaultQueue
from elisa.core.interfaces import IQueue
from elisa.core import common, media_uri
from elisa.core.media_db import DummyRow

from math import sqrt
import random
import datetime, time

class SmartQueue(DefaultQueue):

    __implements__ = IQueue

    name = "smart_queue"

    provides = "player:queue"

    default_config = {
        'update_rating_method': 'conservative', # or 'accelerated'
        'check_rating_method': 'bell', # or 'thresh'
        'repeat_avoid_method': 'count', # or 'age'        
        }

    db_tables = {'media_stats': """
                                id INTEGER PRIMARY KEY,
                                last_played TIMESTAMP,
                                nb_played INT NOT NULL DEFAULT 0,
                                current_age TIMESTAMP,
                                count_to_skip INT NOT NULL DEFAULT 0
                                """}

    def __init__(self):
        DefaultQueue.__init__(self)

        conf = self.get_config()
        self.update_rating_method = conf['update_rating_method']
        self.min_score = 0
        self.max_score = 100
        self.increase_factor = 10.
        self.decrease_factor = 10.
        # bell curve model
        self.bell = [ 0.0000, 0.0987, 0.1915, 0.2734, 0.3413, 0.3944, 0.4332,
                      0.4599, 0.4772, 0.4878, 0.4938, 0.4970, 0.4987, 0.4994,
                      0.4997, 0.4999, 0.5000 ]

        check_rating_method = conf['check_rating_method']
        # use bell method by default
        self.check_rating = self._check_bell
        if check_rating_method == 'thresh':
            self.check_rating = self._check_threshold

        # repeat-avoidance 'count' or 'age'
        repeat_avoid_method = conf['repeat_avoid_method']
        # by default we check by age
        self.check_repeat = self._check_age
        if repeat_avoid_method == 'count':
            self.check_repeat = self._check_count
        self.min_age = 3600 # 1 hour (should be configurable ?)
        self.max_count = 10 #        (idem)
        self.age_bypass_count = 0 # this is set later

    def next_item(self, head_uri=None, current=None, at_user_request=False):
        """ Select the next uri of the queue and return it as a tuple
        (uri, short_name)
        """        
        next_location = None
        short_name = None

        if not current:
            current, short_name = self.get_current_item()

        if current:
            current_media = self.get_media_with_uri(current)
            if current_media:
                if current_media.format == 'audio':
                    if at_user_request:
                        self.decrease_rating(current_media)
                    else:
                        self.increase_rating(current_media)
                    self.update_media_history(current_media)
                else:
                    return (None, None)
                
        good_song = aged_song = False
        medias = self.get_medias()
        if medias:
            stats = self.compute_ratings(medias)
            while (not good_song) or (not aged_song):
                # find a candidate
                random_number = random.randint(0,len(medias)-1)
                candidate = medias[random_number]
                # improve it
                good_song = self.check_rating(candidate, stats)
                if good_song:
                    aged_song = self.check_repeat(candidate)
            media = self.get_media_with_id(candidate.id)
            next_location, short_name = media.uri, media.short_name
        return next_location, short_name

    def compute_ratings(self, medias):
        """ do some stats (used by checkBell & checkThreshold) """
        #ratings = self.data.getRatings()
        ratings = [ r.rating for r  in medias ]
        std_dev_counter = 0
        ratings_number = float(len(ratings))
        #for rating in ratings:
        #    mean_counter += rating
        #mean = mean_counter / ratings_number
        mean = self.sql_execute("select avg(rating) as a from media")[0].a
        for rating in ratings:
            std_dev_counter += (rating - mean)**2
        # standard deviation
        std_dev = sqrt( std_dev_counter / ratings_number)
        threshold = std_dev + mean 
        reprieve_threshold = mean - std_dev
        return {'std_dev': std_dev, 'threshold': threshold, 'mean': mean,
                'reprieve_threshold': reprieve_threshold }

    # --- Song rating stuff
    def increase_rating(self,media):
        """ increase the rating of the current queue media """
        current_rating = media.rating
        if self.update_rating_method == 'conservative':
            increase = int((self.max_score - current_rating) / self.increase_factor)
        elif self.update_rating_method == 'accelerated':
            x = current_rating - self.min_score
            y = self.max_score - current_rating
            increase = int( max( [1 , min([x,y])/2.] ) )
        new_rating = current_rating + increase
        if new_rating > self.max_score:
            new_rating = self.max_score
        self.update_media(media, rating=new_rating)

    def decrease_rating(self, media):
        """ decrease the rating of the current media """
        current_rating = media.rating
        if self.update_rating_method == 'conservative':
            decrease = current_rating / self.decrease_factor
        elif self.update_rating_method == 'accelerated':
            x = current_rating - self.min_score
            y = self.max_score - current_rating
            decrease = int( max( [1 , min([x,y])/2.] ) )
        new_rating = current_rating - decrease
        if new_rating < self.min_score:
            new_rating = self.min_score
        self.update_media(media, rating=new_rating)

    # --- Repeat avoidance methods
    def _check_age(self, media):
        aged = False
        current_age = media.current_age
        if not current_age:
            aged = True
        else:
            now = time.time()
            prev = time.mktime(time.strptime(str(current_age),
                                             '%Y-%m-%d %H:%M:%S'))
            current_age = now - prev

            if (current_age >= self.min_age) or (self.age_bypass_count == 0):
                aged = True
            else:
                self.age_bypass_count -= 1
        return aged
    
    def _check_count(self, media):
        aged = False
        count_to_skip = media.count_to_skip
        if not count_to_skip:
            aged = True
        else:
            current_count = count_to_skip - 1
            if current_count <= 0:
                current_count = 0
                aged = True
            self.update_media(media, count_to_skip=current_count)
        return aged

    # --- Rating based choosing methods
    def _check_bell(self, media, stats):
        good_song = neg_sigma = False
        float_random = random.random()
        if stats['std_dev'] == 0:
            current_sigma = 0
        else:
            pre_rating = media.rating
            current_sigma = (pre_rating - stats['mean']) / stats['std_dev']
        if current_sigma < 0:
            current_sigma = - current_sigma
            neg_sigma = True
        if current_sigma > 4:
            area_sigma = 0.5
        else:
            lower_sigma = int(current_sigma*4)/4.
            upper_sigma = (int(current_sigma*4)/4.) + .25
            lower_area = self.bell[int(current_sigma*4)]
            upper_area = self.bell[int(current_sigma*4)+1]
            ratio_sigma = (upper_sigma - current_sigma) / (upper_sigma - lower_sigma)
            area_sigma = lower_area * ratio_sigma + upper_area * (1 - ratio_sigma)
        if neg_sigma:
            proba = 0.5 - area_sigma / 100.
        else:
            proba = 0.5 + area_sigma / 100.
        if proba > float_random:
            good_song = True
        return good_song

    def _check_threshold(self, media, stats):
        good_song = False
        rating = media.rating
        if rating >= stats['threshold']:
            good_song = True
        elif rating >= stats['reprieve_threshold']:
            rand = random.randint(0,9)
            if rand == 0:
                good_song = True
        return good_song

    # SQL API

    def get_medias(self):
        request = "select media.id, media.rating, media_stats.* from media, "\
                  "media_stats where media.format='audio' and "\
                  "media.id=media_stats.id"
        rows = self.sql_execute(request)
        ids = [r.id for r in rows]
        all_audio = "select id, rating from media where format='audio'"
        for row in self.sql_execute(all_audio):
            if row.id not in ids:
                new_row = DummyRow({'id': row.id, 'rating': row.rating,
                                    'last_played': None, 'nb_played': 0,
                                    'current_age': None, 'count_to_skip': 0
                                    })
                rows.append(new_row)
        return rows
    
    def get_media_with_id(self, media_id):
        request = "select * from media where id=%s" % media_id
        return self.sql_execute(request)[0]

    def get_media_with_uri(self, uri):
        app = common.get_application()
        return app.get_media_manager().get_media_for_uri(uri)

    def update_media(self, media, **kw):
        update_media_req = ""
        new_values = []
        for key, value in kw.iteritems():
            if key == 'rating':
                update_media_req = "update media set rating=%d where id=%s"
                update_media_req = update_media_req % (value, media.id)
            else:
                new_values.append((key,value))
                
        if update_media_req:
            self.sql_execute(update_media_req)

        if new_values:
            req = "select id from media_stats where id=%s" % media.id
            do_insert = len(self.sql_execute(req)) == 0
            if do_insert:
                req = "insert into media_stats(id, %s) values (%s, %s)"
                columns = ','.join([str(k) for k,v in new_values])
                col_values = ','.join([str(v) for k,v in new_values])
                req = req % (columns, media.id, col_values)
            else:
                values = ', '.join(["%s=%s" % (k,v) for k,v in new_values])
                req = "update media_stats set %s where id=%s" % (values,
                                                                 media.id)
            self.sql_execute(req)

    def update_media_history(self, media):
        do_insert = False
        curr_stats = "select nb_played from media_stats where id=%s" % media.id
        result = self.sql_execute(curr_stats)
        if result:
            nb_played = result[0].nb_played + 1
        else:
            nb_played = 1
            do_insert = True

        now = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))
        last_played = now
        current_age = now
        count_to_skip = 10
        
        if do_insert:
            query = "insert into media_stats(id, last_played, nb_played, "\
                    "current_age, count_to_skip) values (%s, %r, %s, %r, %s)"
            query = query % (media.id, last_played, nb_played, current_age,
                             count_to_skip)
        else:
            query = "update media_stats set "
            dict_values = {'last_played': repr(last_played),
                           'current_age': repr(current_age),
                           'count_to_skip': count_to_skip,
                           'nb_played': nb_played
                           }
            values = ','.join(["%s=%s" % (k,v)
                               for k,v in dict_values.iteritems()])
            query += "%s where id=%s" % (values, media.id)
        self.sql_execute(query)
