# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import log
from elisa.core import common
from elisa.core import component
from elisa.extern import msgfmt
from twisted.internet import defer
import os
import gettext
import locale

"""
Borrowed some code from this recipe:
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/204349
"""

class IPlugin(component.Interface):
    """
    This is the base interface for our plugin system. Every Plugin
    sub-class has to implement it.

    You can declare the default methods each plugin should implement here.
    """

class ExtensionPoint(property):

    def __init__(self, interface):
        property.__init__(self)
        self.interface = interface

    def __get__(self, calling_plugin, typ):
        """ Walk through the loaded plugins, find the ones
        implementing our interface, sort them by feature support and
        return the resulting hash.

        """
        if not hasattr(self, '_matched_extensions'):
            self._matched_extensions = {}
            plugin_manager = common.get_plugin_manager()            
            for plugin in plugin_manager.get_singleton_plugins():
                if plugin.implements(self.interface):
                    for feature in plugin.provides:
                        self._matched_extensions[feature] = plugin
            for entry_point_name, plugins in plugin_manager.get_plugins().iteritems():
                plugins = [ plugin for plugin in plugins
                            if plugin.implements(self.interface) ]

                for plugin in plugins:
                    for feature in plugin.provides:
                        if feature not in self._matched_extensions and \
                           plugin.enabled:
                            p = plugin()
                            p.post_init()
                            self._matched_extensions[feature] = p

        extensions = self._matched_extensions
        self._matched_extensions = dict([(f, p)
                                         for (f,p) in extensions.iteritems()
                                         if p.initialized])
        return self._matched_extensions


class Plugin(component.Component):

    "Plugin Base Class"

    default_config = {}

    # configuration doc. keys should be same as the ones in default_config
    # and values should be strings
    config_doc = {}
    
    name = "default"

    logger = log.Logger()

    provides = []

    """
    plugin's .po directory. If not empty, .po files will be compiled and
    localization support will be installed. Plugins need to override local _() function
    in their sources to use localization. Please see docs/plugins.txt
    IMPORTANT: .po file must have the same name as the Plugin's name property
    """
    translation_path = ""

    """
    __platforms__ : empty list means all platforms are supported, supported
    values must be defined in os.name or platform.system()
    """
    __platforms__ = []
        
    singleton = False

    db_tables = {}
    
    def __init__(self):
        """
        Constructor
        """
        component.Component.__init__(self)
        self.load_config()

        # localization support
        if self.translation_path:
            self.install_localization()
            
        self.initialized = True

    def post_init(self):
        for table_name, sql in self.db_tables.iteritems():
            query = "CREATE TABLE %s (%s);" % (table_name, sql)
            self.sql_execute(query)
        
    def unload(self, message=None):
        dfr = defer.Deferred()
        manager = common.get_plugin_manager()
        if manager.is_registered(self.__class__):
            if message:
                self.logger.info(message)
            self.logger.info("Un-loading %s" % self.name)
            self.store_config()
            manager.unregister_plugin(self.__class__)
            self.initialized = False
            # XXX: hrm for some strange reason, wrong plugin is deleted
            #      from singletons list .... weird, it is.
            if self.singleton:
                manager.del_singleton_plugin(self)
        return dfr
    
    def load_config(self):
        """ Load the plugin's configuration. If none found, create it
        using the default config stored in `default_config` Plugin
        attribute.
        """
        self.logger.debug('Plugin %s is loading its config' % self.name)
        self.config = {}
        appli = common.get_application()
        if appli:
            config = appli.get_config()
            section_name = "plugins.%s" % self.get_name()

            section = config.get_section(section_name)
            if not section:
                section = self.default_config
                config.set_section(section_name, self.default_config,
                                   doc=self.config_doc)

            else:
                for option, value in self.default_config.iteritems():
                    if option not in section:
                        section[option] = value

            self.config = section

    def store_config(self):
        appli = common.get_application()
        if appli:

            # warn of un-documented options
            config_option_names = set(self.get_config().keys())
            documented_options = set(self.config_doc.keys())
            intersection = list(config_option_names - documented_options)
            if intersection:
                msg = "Undocumented options for %r: %r" % (self.get_name(),
                                                           intersection)
                self.logger.info(msg)
                                                                      
            section_name = "plugins.%s" % self.get_name()
            appli.get_config().set_section(section_name, self.get_config(),
                                           doc=self.config_doc)

    def get_config(self):
        """ Configuration accessor. Only the plugin's config is
        visible from here."""
        return self.config

    def get_name(self):
        return self.name

    def sql_execute(self, query, log=False):
        result = None
        media_mgr = common.get_application().get_media_manager()
        if media_mgr.enabled():
            connection = media_mgr.get_db_connection()
            result = connection.execute(query, log=log)
        return result

    def install_localization(self):
        for path, names, filenames in os.walk(self.translation_path):
            for i, dirname in enumerate(names):
                if dirname.startswith('.'):
                    del names[i]
                for f in filenames:
                    if f.endswith('.po'):
                        src = os.path.join(path, f)
                        dest = src[:len(src) - 3] + '.mo'
                        # compile the .po if necessary
                        if not os.path.exists(dest):
                            msgfmt.make(src, dest)
                        else:
                            src_mtime = os.stat(src)[8]
                            dest_mtime = os.stat(dest)[8]
                            if src_mtime > dest_mtime:
                                msgfmt.make(src, dest)

    @classmethod
    def enable(cls):
        cls.logger.info('Enabling plugin %s' % cls.name)
        app_config = common.get_application().get_config()
        ep_name = cls.entry_point_name
        option = app_config.get_option(ep_name)
        if type(option) == list:
            option.append(cls.name)
        else:
            option = cls.name
        app_config.set_option(ep_name, option)

    def disable(self):
        self.logger.info('Disabling plugin %s' % self.name)
        app_config = common.get_application().get_config()
        ep_name = self.entry_point_name
        option = app_config.get_option(ep_name)
        if type(option) == list and self.name in option:
            option.remove(self.name)
        else:
            option = ''
        app_config.set_option(ep_name, option)
        

#test
if __name__ == "__main__":
    gettext.bindtextdomain("test", "i18n_test")
    _ = lambda s: gettext.dgettext("test", s);
    
    class TestPlug(Plugin):
        translation_path = "i18n_test"
        name = "test"
        __implements__ = IPlugin
        
        def __init__(self):
            Plugin.__init__(self)
            
        def say_yes(self):
            print _("Yes this is a test")
                
    t = TestPlug()
    t.say_yes()
