# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.plugin import Plugin, IPlugin
from elisa.core import common
from elisa.core.signals.player import stopped_signal
from elisa.core import __version__ as elisa_version

import urllib2,urllib
import re,time,md5, os
import xml.utils.iso8601,  thread, threading
from cPickle import dump, load

from twisted.internet import defer, threads
import gst

UNKNOWN = 'unknown'
SAVED_TRACKS_FILE = os.path.expanduser('~/.elisa/unsubmitted_tracks.txt')


class ScrobbledTrack:
    def __init__(self, media, length):
        if media.artist != 'unknown artist':
            self.artist = media.artist
        else:
            self.artist = UNKNOWN
        if media.album != 'unknown album':
            self.album = media.album
        else:
            self.album = UNKNOWN
        self.name = media.song
        self.length = str(length)

        # MusicBrainz not supported yet
        self.mbid = None
        
        self.date = re.sub("(\d\d\d\d-\d\d-\d\d)T(\d\d:\d\d:\d\d).*","\\1 \\2",
                           xml.utils.iso8601.tostring(time.time()))

    def isSubmitable(self):
        return UNKNOWN not in (self.artist, self.album)

    def __repr__(self):
        return "%r by %r from %r (%r @ %r)" % (self.name, self.artist,
                                               self.album,
                                               self.length, self.date)

    def urlencoded(self, num):
        encode = ""
        encode += "a["+str(num)+"]="+urllib.quote_plus(self.artist)
        encode += "&t["+str(num)+"]="+urllib.quote_plus(self.name)
        encode += "&l["+str(num)+"]="+urllib.quote_plus(self.length)
        encode += "&i["+str(num)+"]="+(self.date)
        if self.mbid is not None:
            encode += "&m["+str(num)+"]="+urllib.quote_plus(self.mbid)
        else:
            encode += "&m["+str(num)+"]="
        encode += "&b["+str(num)+"]="+urllib.quote_plus(self.album)
        return encode


class AudioScrobblerPlugin(Plugin):

    __implements__ = IPlugin

    config_doc = {'user': 'Last.FM username',
                  'password': 'Last.FM password for the user :-)'
                  }

    default_config = {'user': 'fill_me',
                      'password': 'fill_me'
                      }
    name = "audioscrobbler"

    singleton = True

    def __init__(self):
        Plugin.__init__(self)
        self.do_unload = False
        self.load_error = ""

        media_manager = common.get_application().get_media_manager()
        user = self.get_config().get('user')
        if user == 'fill_me':
            self.load_error = "Please configure your Last.FM account settings"
            self.do_unload = True
        elif not media_manager.enabled():
            self.load_error = "Media manager is required for this plugin"
            self.do_unload = True
        else:
            self.url = "http://post.audioscrobbler.com/"
            self.user = user
            self.password = self.get_config().get('password')
            self.client = "eli"
            self.version = "0.1"
            self.lock = threading.Lock()
            self.loadSavedTracks()

            self.handshake()
            stopped_signal.connect(self._player_stopped)

    def post_init(self):
        if self.do_unload:
            self.unload(self.load_error)

    def _player_stopped(self, playable):
        length = playable.get_length() / gst.SECOND
        status = playable.get_status() / gst.SECOND
        uri = playable.get_uri()
        
        if playable.contains_audio() and not playable.contains_video():
            if status > 0 and status == length:
                media_manager = common.get_application().get_media_manager()
                media = media_manager.get_media_for_uri(uri)
                if media:
                    track = ScrobbledTrack(media, length)
                    dfr = threads.deferToThread(self.submit, track)

                    def submitted(result):
                        self.debug("submit result: %s" % result)

                    dfr.addCallback(submitted)

    def loadSavedTracks(self):
        self.tracksToSubmit = []
        if os.path.exists(SAVED_TRACKS_FILE):
            f = open(SAVED_TRACKS_FILE, 'r')
            try:
                self.tracksToSubmit = load(f)
            except:
                pass
            f.close()

    def saveTracks(self):
        # TODO: call this on application.stop()
        self.debug('Saving %s tracks' % len(self.tracksToSubmit))
        f = open(SAVED_TRACKS_FILE, 'w')
        dump(self.tracksToSubmit, f)
        f.close()

    def debug(self, msg):
        self.logger.debug("[AudioScrobbler] %s" % msg)

    def handshake(self):
        self.logged = False
        self.debug("Handshaking...")
        url = self.url+"?"+urllib.urlencode({
            "hs":"true",
            "p":"1.1",
            "c":self.client,
            "v":self.version,
            "u":self.user
            })
        try:
            result = urllib2.urlopen(url).readlines()
        except Exception, ex:
            self.debug(ex)
        else:
            status = result[0]
            if status.startswith("BADUSER"):
                return self.baduser(result[1:])
            if status.startswith("FAILED"):
                return self.failed(result)
            self.logged = True
            if status.startswith("UPTODATE") or status.startswith("UPDATE"):
                return self.uptodate(result[1:])
            return True
        return False

    def uptodate(self, lines):
        self.md5 = re.sub("\n$","", lines[0])
        self.debug("MD5 = %r" % self.md5)
        self.submiturl = re.sub("\n$","", lines[1])
        self.debug("submiturl = %r" % self.submiturl)
        self.interval(lines[2])
        return True

    def baduser(self, lines):
        self.debug("Bad user")
        self.interval(lines[1])
        return False

    def failed(self, lines):
        self.debug('Failed : %s' % lines[0])
        self.interval(lines[1])
        return False

    def interval(self, line):
        match = re.match("INTERVAL (\d+)", line)
        if match is not None:
            secs = int(match.group(1))
            self.debug("Sleeping a while (%s seconds)" % secs)
            time.sleep(secs)

    def submit(self, track):
        if not self.logged:
            # try to log
            if not self.handshake():
                if track not in self.tracksToSubmit:
                    self.debug("Queued submission of track %s" % track)
                    self.tracksToSubmit.append(track)
                self.debug("%s track(s) currently queued" % len(self.tracksToSubmit))
                self.saveTracks()
                return

        if track not in self.tracksToSubmit:
            self.tracksToSubmit.append(track)

        self.debug("Will try to submit tracks : %s" % str(self.tracksToSubmit))

        try:
            md5response = md5.md5(md5.md5(self.password).hexdigest()+self.md5).hexdigest()
            self.debug('md5response: %s' % md5response)
            
            post = "u="+self.user+"&s="+md5response
            count = 0
            self.debug('post: %r' % post)
            
            for track in self.tracksToSubmit:
                l = int(track.length)
                if not track.isSubmitable():
                    self.debug("Missing informations from track, skipping submit")
                    self.debug(track)
                elif l < 30 or l > (30*60):
                    self.debug("Track is too short or too long, skipping submit")
                    self.debug(track)
                else:
                    self.debug('encoded: %r' % track.urlencoded(count))
                    post += "&"
                    post += track.urlencoded(count)
                    count += 1
        except Exception, ex:
            self.debug('Exception : %s' % ex)

        self.debug('count = %s' % count)
        self.debug(post)
        post = unicode(post)
        if count:
            try:
                result = urllib2.urlopen(self.submiturl,post)
            except Exception,ex:
                self.debug(ex)
                self.logged = False
            else:
                results = result.readlines()
                self.debug("submit result : %r" % results)
                if results[0].startswith("OK"):
                    self.interval(results[1])
                    self.tracksToSubmit = []
                    self.saveTracks()
                elif results[0].startswith("FAILED"):
                    self.failed(results)
                    self.handshake()
                elif results[0].startswith("BADAUTH"):
                    self.baduser(results)
                    self.handshake()
