# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.base_plugins import DataAccessPlugin, DataAccessFile
from elisa.core.interfaces import IDataAccess
from elisa.core.utils import misc
from elisa.core import media_uri, common
from elisa.core.signals.data import new_device, del_device

import socket
import pygst
pygst.require('0.10')
import gst
import gobject

try:
    import daap
except ImportError:
    daap = None

try:
    import dbus
    if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
        import dbus.glib
except ImportError:
    dbus = None

if dbus:
    try:
        import avahi
    except ImportError:
        avahi = None
else:
    avahi = None
    
"""
TODO:

- add a "copy to local library" feature?

"""

class DaapSource(gst.BaseSrc):
    __gsttemplates__ = (
        gst.PadTemplate("src",
                        gst.PAD_SRC,
                        gst.PAD_ALWAYS,
                        gst.caps_new_any()),
        )

    __gstdetails__ = ("DAAP plugin", "Foo/Bar", "Read data on DAAP shares",
                      "Philippe Normand <philippe@fluendo.com>")

    blocksize = 4096
    fd = None
    
    def __init__(self, name):
        self.__gobject_init__()
        self.curoffset = 0
        self._libs = {}
        self.cnx = daap.DAAPClient()
        self.set_name(name)
        
    def get_library(self, host, port):
        if (host, port) not in self._libs:
            self.cnx.connect(host, port)
            session = self.cnx.login()
            library = session.library()
            self._libs[(host, port)] = library
        return self._libs.get((host, port))
            
    def set_property(self, name, value):
        if name == 'uri':
            parsed = media_uri.Uri(value).get_parts()
            port = int(parsed['port'] or  '3689')
            library = self.get_library(parsed['host'], port)
            #track_id = int(parsed['path'][1:])
            track_id = int(parsed['params'].get('id','1'))
            track = [ t for t in library.tracks()
                      if t.id == track_id ][0]
            self.fd = track.request()
            #print '%s/%s - %s' % (track.artist, track.album,
            #                      track.name)
            
    def do_create(self, offset, size):
##         if offset != self.curoffset:
##             self.fd.seek(offset, 0)
        data = self.fd.read(self.blocksize)
        if data:
            self.curoffset += len(data)
            return gst.FLOW_OK, gst.Buffer(data)
        else:
            return gst.FLOW_UNEXPECTED, None
        
gobject.type_register(DaapSource)


class AvahiMonitor:

    def __init__(self):
        self._callbacks = {'new-service':  [],
                           'remove-service': []
                           }
        self.bus = dbus.SystemBus()
        avahi_bus = self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER)
        self.server = dbus.Interface(avahi_bus, avahi.DBUS_INTERFACE_SERVER)

        stype = '_daap._tcp'
        domain = 'local'
        avahi_browser = self.server.ServiceBrowserNew(avahi.IF_UNSPEC,
                                                      avahi.PROTO_UNSPEC,
                                                      stype, domain,
                                                      dbus.UInt32(0))
        obj = self.bus.get_object(avahi.DBUS_NAME, avahi_browser)
        self.browser = dbus.Interface(obj, avahi.DBUS_INTERFACE_SERVICE_BROWSER)

    def start(self):
        self.browser.connect_to_signal('ItemNew', self.new_service)
        self.browser.connect_to_signal('ItemRemove', self.remove_service)

    def stop(self):
        self.bus.close()
                
    def new_service(self, interface, protocol, name, type, domain, flags):
        service = self.server.ResolveService(interface, protocol,
                                             name, type, domain,
                                             avahi.PROTO_UNSPEC, dbus.UInt32(0))
        address, port = service[-4:-2]
        name = str(service[2])
        for cb in self._callbacks['new-service']:
            cb(name, address, port)
            
    def remove_service(self, interface, protocol, name, type, domain):
        service = self.server.ResolveService(interface, protocol,
                                             name, type, domain,
                                             avahi.PROTO_UNSPEC, dbus.UInt32(0))
        address, port = service[-4:-2]
        name = str(service[2])
        for cb in self._callbacks['remove-service']:
            cb(name, address, port)
            
        
    def add_callback(self, sig_name, callback):
        self._callbacks[sig_name].append(callback)

    def remove_callback(self, sig_name, callback):
        self._callback[sig_name].remove(callback)


class DaapFSPlugin(DataAccessPlugin):

    __implements__ = IDataAccess

    name = 'daap_fs'
    
    provides = ["uri:daap",]
    singleton = True

    default_config = {}

    root_folder_icon = 'folder-music.png'
    uncacheable_uri_schemes = ['daap']

    def __init__(self):
        DataAccessPlugin.__init__(self)
        self._libraries = {}
        self._monitor = None
        self.gst_pipeline = None
        if not daap:
            url = "http://jerakeen.org/code/PythonDaap/"
            self.logger.info("PythonDAAP not found. Check it out at %s" % url)
            self.unload()
        if not dbus:
            if not avahi:
                self.logger.info("Please install python-avahi to get Avahi support")
            self.logger.info("python-dbus is missing...")
            self.logger.info("Avahi support disabled. DAAP shares won't appear automagically")            
        else:
            try:
                self._monitor = AvahiMonitor()
                self._monitor.add_callback('new-service', self._add_location)
                self._monitor.add_callback('remove-service',
                                           self._remove_location)
            except Exception, ex:
                self.logger.info("Couln't initialize Avahi monitor: %s" % str(ex))
                self.unload()
            else:
                self._monitor.start()

##     def unload(self):
##         if self._monitor:
##             self._monitor.stop()
##         DataAccessPlugin.unload(self)

    def get_gst_pipeline(self):
        if not self.gst_pipeline:            
            self.gst_pipeline = gst.Pipeline('pipeline')

            daapsrc = DaapSource('src')
            decoder = gst.element_factory_make('decodebin')
            convert = gst.element_factory_make('audioconvert')

            def on_new_decoded_pad(element, pad, last):
                caps = pad.get_caps()
                name = caps[0].get_name()
                apad = convert.get_pad('sink')
                if 'audio' in name:
                    if not apad.is_linked(): # Only link once
                        pad.link(apad)

            decoder.connect('new-decoded-pad', on_new_decoded_pad)

            
            volume = gst.element_factory_make('volume', 'vlm')
            sink = gst.element_factory_make('autoaudiosink', 'audio-sink')

            self.gst_pipeline.add(daapsrc, decoder, convert, volume, sink)

            daapsrc.link(decoder)
            convert.link(volume)
            volume.link(sink)
            
        return self.gst_pipeline
        
    def _add_location(self, name, address, port):
        name = 'DAAP: %s' % name
        self.logger.info("New DAAP share at %s:%s : %s" % (address, port, name))
        address = misc.fs_encode(address)
        uri = 'daap://%s:%s/' % (address, port)
        new_device.emit(name, uri, 'music')
        
    def _remove_location(self, name, address, port):
        self.logger.info("DAAP share at %s:%s disappeared" % (address, port))
        address = misc.fs_encode(address)
        uri = 'daap://%s:%s/' % (address, port)
        del_device.emit(name, uri, 'music')
        
    def _get_library(self, uri):
        parsed_uri = self._parse(uri)

        host = parsed_uri.host
        if parsed_uri.port:
            port = int(parsed_uri.port)
        else:
            port = 3689

        address = (host, port)
        if address not in self._libraries:
            connection  = daap.DAAPClient()
            try:
                connection.connect(*address)
            except Exception, exc:
                self.logger.info("Couldn't connect: %s" % str(exc))
            else:
                # auth isn't supported yet. Just log in
                session = connection.login()
            
                library = session.library()

                try:
                    self._libraries[address] = self._build_library(library.tracks())
                except (socket.timeout, socket.error):
                    pass
                
        return self._libraries.get(address)
        
    def _build_library(self, tracks):
        lib = {}
        for track in tracks:
            if None in (track.artist, track.album):
                continue
            if not lib.get(track.artist):
                lib[track.artist] = {}
            if not lib[track.artist].get(track.album):
                lib[track.artist][track.album] = {}
            if not lib[track.artist][track.album].get(track.id):
                lib[track.artist][track.album][track.id] = (track.name,
                                                            )
        return lib
                            
    def is_dir(self, uri):
        uri = self._parse(uri)
        return uri.scheme == 'daap' and uri.params == {}

    def _parse(self, uri):
        uri = misc.fs_encode(uri)
        uri = media_uri.Uri(location=uri, quote=True)
        return uri

    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        file_type = ''
        return file_type

    def get_direct_children(self, uri):
        children = {}
        tracks = []
        parsed_uri = self._parse(uri)

        lib = self._get_library(uri)
        if not lib:
            return {}
        
        path = parsed_uri.path
        if path == '/':
            # fetch all artists and make some folders
            index = 0
            for artist in lib.keys():
                child_uri = parsed_uri.copy()
                artist = misc.fs_encode(artist)
                
                child_uri.path = '/%s' % artist
                child_type = 'directory'
                file_type = ''
                children[str(child_uri)] = (child_type, file_type, artist,
                                            index)
                index += 1
        elif parsed_uri.params.get('id'):
            return {}
        else:
            if path.endswith('/'):
                path = path[:-1]
            path = path[1:].split('/')
            track_id = None
            if len(path) == 2:
                # artist/album
                artist, album = path
                artist = unicode(artist)
                album = unicode(album)
                index = 0
                for track_id, track in lib.get(artist,{}).get(album,{}).iteritems():
                    child_uri = parsed_uri.copy()
                    # TODO: make an URI easily playable by GStreamer...
                    child_uri.path = '/%s/%s?id=%s' % (misc.fs_encode(artist),
                                                       misc.fs_encode(album),
                                                       track_id)
                    child_type = 'file'
                    file_type = 'audio'
                    track_title = track[0]
                    children[str(child_uri)] = (child_type, file_type,
                                                track_title, index)
                    index += 1
            else:
                # artist/
                artist = path[0]
                index = 0
                for album_name, album in lib.get(artist,{}).iteritems():
                    child_uri = parsed_uri.copy()
                    child_uri.path = '/%s/%s' % (misc.fs_encode(artist),
                                                 misc.fs_encode(album_name))
                    child_type = 'directory'
                    file_type = ''
                    children[str(child_uri)] = (child_type, file_type,
                                                album_name, index)
                    index += 1
                
        return children

    def has_children(self, uri):
        return True

    def next_location(self, uri, root=None):
        """
        Return the uri just next to `from_uri`.
        """
        pass
    
    def open(self, uri, mode=None):
        """
        Open an uri and return DataAccessFile file
        """
        pass
    
    def remove_location(self, uri):
        """
        Delete data at location identified by its uri.
        """
        pass


if __name__ == '__main__':
    # FIXME: this autotest won't work anymore because media manager is now part
    #        of Application
    try:
        from twisted.internet import glib2reactor
        glib2reactor.install()
    except AssertionError:
        pass

    from elisa.core.plugin_manager import PluginManager
    from elisa.core.cache_manager import CacheManager
    from elisa.core.config import Config
    from elisa.core import common
    
    from twisted.internet import reactor
    
    conf = Config("../../local.conf")
    manager = PluginManager(conf)
    common._set_plugin_manager(manager)
    cache = CacheManager(conf)
    common._set_cache_manager(cache)
    manager.load_plugins(False)

    def test():
        f = DaapFSPlugin()
        f.post_init()
        children = f.get_direct_children
        #print children('daap://ix/')
        print children('daap://ix/Radiohead/')
        print children('daap://ix/Radiohead/Amnesiac')
        print children('daap://192.168.1.233/Katie Melua/Piece By Piece')
        reactor.stop()
        
    reactor.callLater(0.4, test)    
    reactor.run()
