# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.base_plugins import DataAccessPlugin, DataAccessFile
from elisa.core.interfaces import IDataAccess
from elisa.core import media_uri
from elisa.core.utils import misc
import copy
import os

"""
TODO:

- support artwork
- support photos (experimental in gpod)
- support videos (?)

"""

try:
    import gpod
except ImportError:
    gpod = None

class Artist:

    def __init__(self, name):
        self.name = name
        self.albums = {}
        
    def add_album(self, album):
        if album.name:
            self.albums[album.name] = album
        
class Album:

    def __init__(self, name):
        self.name = name
        self.songs = {}

    def add_song(self, song):
        self.songs[song.title] = song

class Song:

    def __init__(self, nr, name, path, typ):
        self.nr = nr
        self.name = name
        self.title = '%s - %s' % (nr, name)
        self.path = path
        self.filetype = typ

def load_artists(db):
    artists = {}

    for track in db:
        artist_name = track['artist']
        album_name = track['album']
        title = track['title']
        track_nb = str(track['track_nr']).rjust(2, '0')
        track_type = track['filetype']
        song_path = track['ipod_path'].replace(':','/')
        song_title = '%s - %s' % (track_nb, title)
        
        
        if artist_name not in artists:
            artist = Artist(artist_name)
            artists[artist_name] = artist
        else:
            artist = artists[artist_name]

        if album_name not in artist.albums:
            album = Album(album_name)
            artist.add_album(album)
        else:
            album = artist.albums[album_name]

        if song_title not in album.songs:
            song = Song(track_nb, title, song_path, track_type)
            album.add_song(song)

    return artists
    

class IPodPlugin(DataAccessPlugin):

    name = "ipod"

    __implements__ = IDataAccess
    __platforms__ = ['posix']

    provides = ["uri:ipod",]

    singleton = True
    root_folder_icon = 'ipod.png'

    uncacheable_uri_schemes = ['ipod']
    
    def __init__(self):
        DataAccessPlugin.__init__(self)
        self.dbs = {}

    def post_init(self):
        if not gpod:
            self.logger.info("libgpod not found. Check out CVS snapshot at http://www.gtkpod.org/libgpod.html")
            self.unload()
        DataAccessPlugin.post_init(self)
        
    def is_dir(self, uri):
        uri = self._parse(uri)
        return uri.scheme == 'ipod'

    def _parse(self, uri):
        uri = misc.fs_encode(uri)
        uri = media_uri.Uri(location=uri, quote=True)
        return uri

    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        file_type = ''
        return file_type

    def get_direct_children(self, uri):
        children = {}
        index = 0

        parsed_uri = self._parse(uri)
        
        params = parsed_uri.params
        artist = params.get('artist')
        album = params.get('album')

        path = misc.fs_encode(parsed_uri.path)
        if path not in self.dbs:
            # lazily fetch ipod db from dbs dict
            try:
                if os.path.isdir(path):
                    db = gpod.Database(path)
                elif os.path.isfile(path): #it's a iTunes database
                    db = gpod.Database('', 1, path)
                else:
                    return {}
            except gpod.DatabaseException:
                return {}
            else:
                artists = load_artists(db)
                self.dbs[path] = (db, artists)
                
        db, artists = self.dbs[path]

        db_artist = None
        db_album = None
        
        if artist:
            db_artist = artists.get(artist)
            
        if album and db_artist:
            db_album = db_artist.albums.get(album)

        if db_album:
            # fetch songs of given album
            for song_title, song in db_album.songs.iteritems():
                child_uri = parsed_uri.copy()
                child_uri.scheme = 'file'
                child_uri.params = ''
                child_uri.path += '/' + song.path
                child_type = 'file'
                #file_type = 'audio'
                file_type = self._get_track_type(song)
                children[str(child_uri)] = (child_type, file_type, song_title,
                                            index)
                index += 1

        elif db_artist:
            # fetch albums list of given artist
            for album_title, album in db_artist.albums.iteritems():
                child_uri = parsed_uri.copy()
                child_uri.params['album'] = album_title
                child_type = 'directory'
                file_type = ''
                children[str(child_uri)] = (child_type, file_type, album_title,
                                            index)
                index += 1
        else:
            # return artists list
            for artist_name, artist in artists.iteritems():
                child_uri = parsed_uri.copy()
                child_uri.params['artist'] = artist_name
                child_type = 'directory'
                file_type = ''
                children[str(child_uri)] = (child_type, file_type, artist_name,
                                            index)
                index += 1
                
        return children

    def _get_track_type(self, track):
        track_type = track.filetype
        typ = ''
        for match in ('audio','video'):
            if match in track_type:
                typ = match
                break
        return typ
            

    def next_location(self, uri):
        """
        Return the uri just next to `from_uri`.
        """

    def open(self, uri, mode=None):
        """
        Open an uri and return DataAccessFile file
        """

    def remove_location(self, uri):
        """
        Delete data at location identified by its uri.
        """
        self.logger.info("Removing %s" % str(uri))


if __name__ == '__main__':
    # FIXME: this autotest won't work anymore because media manager is now part
    #        of Application
    from elisa.core.plugin_manager import PluginManager
    from elisa.core.cache_manager import CacheManager
    from elisa.core.config import Config
    from elisa.core import common
    from twisted.internet import reactor

    conf = Config("local.conf")
    manager = PluginManager(conf)
    common._set_plugin_manager(manager)
    cache = CacheManager(conf)
    common._set_cache_manager(cache)
    manager.load_plugins(False)

    def test():
        f = IPodPlugin()
        f.post_init()
        children = f.get_direct_children
        print children('ipod:///media/ipod')
        print children('ipod:///media/ipod?artist=Venus')
        print children('ipod:///media/ipod?album=Garden%20Ruin&artist=Calexico')
        reactor.stop()
        
    reactor.callLater(0.4,test)
    reactor.run()
