# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.plugin import IPlugin
from elisa.core.base_plugins import TreePlugin
from elisa.core.menu import MenuItem
from elisa.core import common, media_uri
import elisa.core.utils.misc
from elisa.core import task as core_task
from elisa.core import log
from elisa.core import plugin
from elisa.core.signals.misc import state_changed
from elisa.core.signals.application import application_started
from elisa.core.utils import image
from elisa.core.interfaces import IDataAccess

from pgm import render

from twisted.internet import task as twisted_task
import Image
import mimetypes
import sys

from PIL import Image as pil

class Slideshow(core_task.Task):

    def __init__(self, loop, slideshow_item,
                 head_uri, interval, pause=None, resume=None, abort=None):
        core_task.Task.__init__(self, slideshow_item,
                                pause=pause, resume=resume,
                                abort=abort)
        self.logger = log.Logger()
        self.head_uri = head_uri
        self.loop = loop
        self.current_uri = head_uri
        self.interval = interval

    def get_head_uri(self):
        return self.head_uri

    def get_current_uri(self):
        return self.current_uri

    def set_current_uri(self, uri):
        self.current_uri = uri

    def running(self):
        return self.loop.running
    
    def pause_slideshow(self):
        if self.loop.running:
            self.loop.stop()
            label = self.get_menu_item().get_short_name()
            self.logger.info("Slideshow %r is paused" % label)

    def resume_slideshow(self):
        if not self.loop.running:
            self.loop.start(self.interval, now=True)
            label = self.get_menu_item().get_short_name()
            self.logger.info("Slideshow %r is resumed" % label)

    def pause_resume(self):
        if self.loop.running:
            self.loop.stop()
            self.loop.start(self.interval, now=False)
        
    def abort_slideshow(self):
        try:
            self.loop.stop()
        except AssertionError:
            pass
        else:
            self.logger.info("Slideshow stopped")
            appli = common.get_application()
            appli.get_task_manager().del_task(self)
            skin = appli.get_skin()
            skin.show_menu()
            skin.reset_background()
            
class PicturesTreePlugin(TreePlugin):
    """
    pictures plugin_tree
    """

    menu_name = _("Pictures")
    menu_help_string = _("Look at pictures")
    __implements__ = IPlugin

    name = "pictures"

    config_doc = {'locations': 'pictures source locations',
                  'slideshow_time_range': 'time in seconds to wait between'\
                                          ' 2 pictures during a slideshow'
                  }
    
    default_config = {'locations': ['file://./sample_data/pictures/*',],
                      'slideshow_time_range': 5}

    provides = ["image"]


    preferred_formats = ["jpeg"]

    item_icon_path = 'unknown.png'
    folder_icon_path = 'folder-pictures.png'

    _current_item = None

    data_access = plugin.ExtensionPoint(IDataAccess)

    def __init__(self):
        if sys.platform == "win32": #Add My Pictures to the default config
            from win32com.shell import shell
            clsid_mypictures = 39
            try:
                p = shell.SHGetSpecialFolderPath(0, clsid_mypictures)
                print "My Pictures directory is:", p
                if p:
                    p = media_uri.Uri(p)
                    self.default_config['locations'].append(p.get_location())
            except:
                print "Could not retrieve My Pictures directory"

        TreePlugin.__init__(self)
        self.key_actions = {}
        application_started.connect(self.application_started)

    def item_filter(self, uri):
        uri = media_uri.Uri(location=uri)
        is_item = False
        data_plugin = self.get_data_plugin(uri)
        if data_plugin:
            handle = data_plugin.open(str(uri))
            if handle:
                if data_plugin.get_name() in ('vfs', 'flickr'):
                    # skip flickr and vfs data reading
                    is_item = True
                else:
                    try:
                        mimetype = mimetypes.guess_type(str(uri))[0].split('/')[0]
                    except:
                        mimetype = 'image'
                    if mimetype == 'image':
                        data = handle.read()
                        try:
                            img = Image.fromstring('RGB', (1, 1), data)
                        except (IOError,ValueError), error:
                            is_item = False
                        else:
                            is_item = True
                data_plugin.close(handle)
        return is_item
                    
    def application_started(self):
        pgmrender = render.get_renderer()

        win = common.get_application().get_window()
        if win:
            win.sig_key_press_event.connect(self.key_pressed)
            self.key_actions = {
                pgmrender.KEY_RIGHT: ('user_next_picture',()),
                pgmrender.KEY_LEFT: ('user_previous_picture', ()),
                }
        else:
            self.logger.info("No skin found, disabling")
            self.unload()
            
    def key_pressed(self, key_event):
        keyval = key_event.keyval
        appli = common.get_application()
        skin = appli.get_skin()
        player = appli.get_player()
        if self.key_actions.has_key(keyval) and not skin.menu_visible() and \
               not skin.dock_visible() and not player.get_playing_uri():
            action, params = self.key_actions.get(keyval)
            if action:
                getattr(self, action)(*params)

    def get_slideshows(self):
        appli = common.get_application()
        return appli.get_task_manager().get_tasks_with_type(Slideshow)
        
    def folder_action_menu(self, menu_item):

        slideshow_icon = 'slideshow.png'
        
        slideshow = MenuItem(short_name=_("Slideshow"))
        slideshow.set_target_path("slideshow")
        slideshow.set_icon_path(slideshow_icon)
        slideshow.set_action_callback(self.start_slideshow_parent)

        menu_item.insert_item(0, slideshow)

    def get_folder_thumbnail(self, uri, widget_item, size):
        # Retrieve max_children thumbnails of pictures contained in uri
        access_plugin = self.data_access.get("uri:%s" % uri.scheme)
        if not access_plugin.get_name() == 'local_fs':
            return

        children_uri = []
        max_children = 4 
        cur = 0
        root = str(uri)
        while cur < max_children:
            uri_str = access_plugin.next_location(str(uri),root=root)[0]
            if not uri_str:
                break
            # skip if child is a directory
            # FIXME: should also check that it is an actual picture
            if access_plugin.get_file_type(uri_str) == '':
                continue
            uri = media_uri.Uri(uri_str)
            children_uri.append(uri)
            cur += 1

        nb_children = len(children_uri)
        if nb_children == 0:
            return

        # box where the thumbnails will be drawn
        ratio = 256.0/size
        padding = (int(40/ratio), int(80/ratio), int(25/ratio), int(60/ratio))
        drawing_box = (padding[0], padding[1], size-padding[2], size-padding[3])
        drawing_box_size = (drawing_box[2] - drawing_box[0],
                            drawing_box[3] - drawing_box[1])
        space = int(10/ratio)

        selected_children = []
        landscape_children = []
        portrait_children = []

        th_manager = common.get_application().get_thumbnail_manager()
        for uri in children_uri: 
            def img_thumbnailer(uri, size):
                thumbnail = image.Image(uri)
                thumbnail.resize(size)
                return thumbnail

            try:
                child_thumbnail = th_manager.get_thumbnail(uri, size, img_thumbnailer)
            except:
                continue

            if child_thumbnail.size[0] > child_thumbnail.size[1]:
                # landscape picture
                child_thumbnail.resize(max(drawing_box_size)/2 - space)
                landscape_children.append(child_thumbnail)
            else:
                # portrait picture
                child_thumbnail.resize(drawing_box_size[1])
                portrait_children.append(child_thumbnail)
        
        selected_children = portrait_children + landscape_children

        # Composite the nb_children thumbnails retrieved using PIL
        #result = pil.new("RGBA", (size, size))
        #result = pil.new("RGBA", (size, size), (255, 255, 255, 255))
        appli = common.get_application()
        folder_path = appli.get_skin().get_picture_path("folder.png")
        result = pil.open(folder_path)

        """
        # cascade
        cur = 0
        for child_thumbnail in selected_children:
            img = pil.frombuffer("RGBA", child_thumbnail.size, child_thumbnail.pixels,
                                    "raw", "RGBA", 0, 1)
            dx = cur*(drawing_box_size[0]-img.size[0])/(nb_children-1) + drawing_box[0]
            dy = cur*(drawing_box_size[1]-img.size[1])/(nb_children-1) + drawing_box[1]
            result.paste(img, (dx, dy), img)
            cur += 1
        """
        
        if len(portrait_children) == 0:
            # mosaique of landscapes
            mvt = [(0,0), (0,1), (1,0), (0,-1)]
        elif len(portrait_children) == 1:
            # mosaique of landscapes and 1 portrait
            mvt = [(0,0), (1,0), (0,1)]
        elif len(portrait_children) >= 2:
            # mosaique of portraits
            mvt = [(0,0), (1,0)]

        dx = drawing_box[0]
        dy = drawing_box[1]
        prec_img_size = (0, 0)
        cur = 0
        for child_thumbnail in selected_children:
            dx += mvt[cur][0]*(prec_img_size[0] + space)
            dy += mvt[cur][1]*(prec_img_size[1] + space)

            img = pil.frombuffer("RGBA", child_thumbnail.size, child_thumbnail.pixels,
                                    "raw", "RGBA", 0, 1)
            result.paste(img, (dx, dy), img)

            prec_img_size = img.size
            cur += 1
            
            if cur >= len(mvt):
                break
 

        pixels = result.tostring('raw', result.mode, 0, 1)
        thumbnail = image.Image(pixels=pixels, size=result.size, format=result.mode)

        widget_item.set_background_image(thumbnail)

    #from elisa.core.utils import profiling
    #@profiling.profileit(4)
    def get_item_thumbnail(self, uri, widget_item, size):
        def img_thumbnailer(uri, size):
            thumbnail = image.Image(uri)
            thumbnail.resize(size)
            return thumbnail

        th_manager = common.get_application().get_thumbnail_manager()
        try:
            thumbnail = th_manager.get_thumbnail(uri, size, img_thumbnailer)
            widget_item.set_background_image(thumbnail)
        except Exception, e:
            self.logger.info("Thumbnail generation failed for %s: %s:" % (uri, str(e)))

    def item_action_menu(self, menu_item):
        pass
##         red_eye_icon = 'slideshow.png'
##         red_eye = MenuItem(short_name=_("Red eyes"))
##         red_eye.set_target_path("red_eye")
##         red_eye.set_icon_path(red_eye_icon)
##         red_eye.set_action_callback(self.red_eye)

##         rotate_icon = 'rotate.png'
##         rotate = MenuItem(short_name=_("Rotate"))
##         rotate.set_target_path("rotate")
##         rotate.set_icon_path(rotate_icon)
##         rotate.set_action_callback(self.rotate)

##         menu_item.add_item(rotate)

##         TreePlugin.item_action_menu(self, menu_item)

##         location = menu_item.get_icon_path()
##         data_plugin = self.get_data_plugin(location)
##         if data_plugin:
##             handle = data_plugin.open(location, 'a+')
##             if not handle:
##                 red_eye.disable()
##                 rotate.disable()
##                 remove_item = menu_item.get_item_with_target('remove')
##                 if remove_item:
##                     remove_item.disable()
##             else:
##                 data_plugin.close(handle)
        
        #menu_item.add_item(red_eye)

    def get_slideshow(self, picture_path):
        picture_uri = media_uri.Uri(picture_path)
        for slideshow in self.get_slideshows():
            root_uri = media_uri.Uri(slideshow.get_head_uri())
            if picture_uri.is_child(root_uri):
                return slideshow
        return None

    def show_picture(self, menu_item, in_slideshow=False):
        self._current_item = menu_item
        target_path = menu_item.get_target_path()
        slideshow = self.get_slideshow(target_path)

        if slideshow and not in_slideshow:
            slideshow.abort_slideshow()
        elif not in_slideshow:
            for slideshow in self.get_slideshows():
                slideshow.abort_slideshow()
                        
        appli = common.get_application()

        self.logger.info('Showing picture %s' % target_path)
        try:
            appli.load_image_uri(target_path, in_slideshow=in_slideshow)
        except Exception, error:
            self.logger.info("Unable to display picture %s: %s" \
                            % (target_path, str(error)))
            raise
        if in_slideshow:
            stats = (media_uri.Uri(target_path).get_label(),'')
            state_changed.emit(None, 'player-stats', stats)
            
    item_action = show_picture

    def start_slideshow_parent(self, menu_item):
        self.start_slideshow(menu_item.get_parent())

    def start_slideshow(self, folder):
        for slideshow in self.get_slideshows():
            slideshow.pause_slideshow()
            
        root = folder.get_target_path()
        self.logger.info("Starting slideshow on %s" % root)
        time_range = self.get_config().get('slideshow_time_range',5)
        slideshow_loop =  twisted_task.LoopingCall(self.next_picture,
                                                   root=root)

        label = folder.get_short_name()
        slideshow_item = MenuItem(label)
        slideshow_item.set_help_string("Slideshow on %s" % label)
        slideshow_item.set_icon_path('slideshow.png')
        slideshow_item.set_target_path(root)
        slideshow_item.set_action_callback(self.toggle_slideshow_pause)
        
        slideshow_task = Slideshow(slideshow_loop, slideshow_item,
                                   root, time_range,
                                   pause=self.pause_slideshow,
                                   resume=self.resume_slideshow,
                                   abort=self.abort_slideshow
                                   )
        
        appli = common.get_application()
        appli.get_skin().hide_menu()
        
        appli.get_task_manager().add_task(slideshow_task)
        slideshow_loop.start(time_range, now=True)

    folder_action = start_slideshow
            
    def toggle_slideshow_pause(self, menu_item):
        
        # pause all
        for slideshow in self.get_slideshows():
            slideshow.pause_slideshow()
            
        # pause/resume currently selected
        slideshow = self.get_slideshow(menu_item.get_target_path())
        if loop.running:
            slideshow.pause_slideshow()
        else:
            slideshow.resume_slideshow()
            
    def red_eye(self, menu_item):
        picture = menu_item.get_parent()
        self.logger.info("Red eyes removal on %s" % picture.get_target_path())

    def rotate(self, menu_item):
        appli = common.get_application()
        picture = menu_item.get_parent()
        appli.get_skin().rotate_image(picture.get_target_path(), -90)
        self.logger.info("Rotate on %s" % picture.get_target_path())

    def next_picture(self, root):
        slideshow = self.get_slideshow(root)
        uri = slideshow.get_current_uri()
        data_plugin = self.get_data_plugin(uri)
        appli = common.get_application()
        if data_plugin:
            next_uri, short_name = data_plugin.next_location(uri, root=root)
            if next_uri:
                is_image = self.item_filter(next_uri)
                while not is_image:
                    next_uri, short_name = data_plugin.next_location(next_uri,
                                                                     root=root)
                    if not next_uri:
                        slideshow.abort_slideshow()
                        return
                    
                    is_image = self.item_filter(next_uri)

                slideshow.set_current_uri(next_uri)

                # show picture!
                create_menu = data_plugin.create_menu_for_uri
                item = create_menu(next_uri, self,
                                   short_name=short_name)
                if item:
                    try:
                        self.show_picture(item, True)
                    except:
                        self.next_picture(root)
                
            else:
                slideshow.abort_slideshow()
                
    def previous_picture(self, root):
        slideshow = self.get_slideshow(root)
        uri = slideshow.get_current_uri()
        data_plugin = self.get_data_plugin(uri)
        appli = common.get_application()
        if data_plugin:
            previous_uri, short_name = data_plugin.previous_location(root)
            if previous_uri:
                slideshow.set_current_uri(previous_uri)

                # show picture!
                create_menu = data_plugin.create_menu_for_uri
                item = create_menu(previous_uri, self,
                                   short_name=short_name)                
                try:
                    self.show_picture(item, True)
                except:
                    self.previous_picture(root)
        
    def user_next_picture(self):        
        for slideshow in self.get_slideshows():
            if slideshow.running():
                self.next_picture(slideshow.get_head_uri())
                slideshow.pause_resume()
                return

        if self._current_item:
            uri = self._current_item.get_target_path()
            data_plugin = self.get_data_plugin(uri)
            appli = common.get_application()
            if data_plugin:
                parent = self._current_item.get_parent()
                if parent:
                    root = parent.get_target_path()
                    next_uri, short_name = data_plugin.next_location(uri,
                                                                     root=root)
                    if next_uri:
                        create_menu = data_plugin.create_menu_for_uri
                        item = create_menu(next_uri, self,
                                           short_name=short_name)
                        if item:
                            item.set_parent(self._current_item.get_parent())
                        # FIXME: checking target_path until skin has
                        #        "empty menu" support
                        if item and item.get_target_path():
                            self._current_item = item
                            try:
                                self.show_picture(item)
                            except Exception, error:
                                self.user_next_picture()
            
    def user_previous_picture(self):
        for slideshow in self.get_slideshows():
            if slideshow.running():
                self.previous_picture(slideshow.get_head_uri())
                slideshow.pause_resume()
                return
        if self._current_item:
            uri = self._current_item.get_target_path()
            data_plugin = self.get_data_plugin(uri)
            appli = common.get_application()
            if data_plugin:
                root = self._current_item.get_parent().get_target_path()
                previous_uri, short_name = data_plugin.previous_location(root)
                if previous_uri:
                    create_menu = data_plugin.create_menu_for_uri
                    item = create_menu(previous_uri, self,
                                       short_name=short_name)
                    if item:
                        item.set_parent(self._current_item.get_parent())
                    # FIXME: checking target_path until skin has
                    #        "empty menu" support
                    if item and item.get_target_path():
                        self._current_item = item
                        try:
                            self.show_picture(item)
                        except:
                            self.user_previous_picture()

    
    def pause_slideshow(self, menu_item):
        parent_uri = menu_item.get_parent().get_target_path()
        slideshow = self.get_slideshow(parent_uri)
        slideshow.pause_slideshow()

    def resume_slideshow(self, menu_item):
        parent_uri = menu_item.get_parent().get_target_path()
        slideshow = self.get_slideshow(parent_uri)
        slideshow.resume_slideshow()

    def abort_slideshow(self, menu_item):
        parent_uri = menu_item.get_parent().get_target_path()
        slideshow = self.get_slideshow(parent_uri)
        slideshow.abort_slideshow()
