# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pgm import widget, render
from pgm.layout import group
from pgm.form import rectangle, label, stroke, color, image
from pgm.timing import controller, modifier, keyframe

from elisa.core.signals.misc import state_changed
from elisa.core.signals.player import settings_signal, playing_signal, stopped_signal
from elisa.core.application import application_started
from elisa.core import common, menu_manager, menu

from elisa.core.widgets import widget_level, progress_bar

from twisted.internet import reactor
import time
import os.path

class ControlMenu(widget_level.WidgetLevel):

    # FIXME: Should not reuse menu_item controls !
    def __init__(self, dock, menu_item):
        width = dock.get_skin().get_window().canvas_size[0]
        self._dock = dock
        self._menu_item = menu_item
        widget_level.WidgetLevel.__init__(self, visible_item_count=10,
                                          arrow_image=None,
                                          width=width)
        self._space = 4
        
##     def width__set(self, width):
##         widget_level.WidgetLevel.width__set(self, width)
##         y = self.y
##         self.first_draw_level()
##         print self.y, y
##         self.y = y
        
    def get_dock(self):
        return self._dock

    def update(self):
        skin = self.get_dock().get_skin()
        playing_item = self._menu_item
        if not self._menu_item:
            self._menu_item = self._dock.menuitem
        if self._menu_item:
            for menu_item in self._menu_item._items[:-1]:
                path = menu_item.get_icon_path()
                path = skin.get_picture_path(path)
                picture = image.Image(path)
                self.add_item(picture, menu_item=menu_item)
            self.first_draw_level()

    def get_widget_item_for_menu_item(self, menu_item):
        for witem in self.get_widget_items(0, self._visible_item_count + 1):
            if witem.menu_item.get_target_path() == menu_item.get_target_path():
                return witem
        return None

    def add_item(self, picture, menu_item=None):
        witem = widget_level.WidgetLevel.add_item(self, picture)
        win = self._dock.get_window()
        buttons_count = float(10)
        size = win.canvas_size[0] / buttons_count
        if size > 90:
            size = 90
        witem.size = (size, size)
        
        if menu_item:
            witem.menu_item = menu_item
        return witem

    def activate_current(self):
        selected_item = self.get_widget_item(self._active_item_index)
        selected_item.menu_item.fire_action()
        #self.unlock_animation()
        
    def _horiz_anim(self, direction):
        selected_item = self.get_widget_item(self._active_item_index)
        if selected_item:
            selected_item.opacity = self._unselected_opacity
            self._anim_end(None)

    def _update_active_list(self):
        pass

    def _anim_end(self, ctrl):
        self.anim_end(self._anim_in_progress)
        self._update_active_list()
        witem = self.get_widget_item(self._active_item_index)
        if witem:
            witem.opacity = 255
        self.unlock_animation()

    def move_right(self): 
        if self.lock_animation():
            if self._active_item_index < self.get_widget_count()-1:
                witem = self.get_widget_item(self._active_item_index)
                
                witem.opacity = self._unselected_opacity
                self._anim_in_progress = 'right'
                self._active_item_index +=1    
                self._horiz_anim('right')
            else:
                self.unlock_animation()
                
    def move_left(self):
        if self.lock_animation():
            if self._active_item_index > 0:
                witem = self.get_widget_item(self._active_item_index)

                witem.opacity = self._unselected_opacity
                self._anim_in_progress = 'left'
                self._active_item_index -=1
                self._horiz_anim('left')
            else:
                self.unlock_animation()
    
class PlayerDock(group.Group):

    def __init__(self, size=(100, 100)):
        self._skin = None
        self._timed_out = None
        self.hide_timeout = 3
        self.loop = False
        application_started.connect(self._application_started)
        stopped_signal.connect(self._player_stopped)
        
        # human speeds
        self._human_speed = {}
        for i in range(-10, 11):
            self._human_speed[str(i)] = 'speed %dX' % i
        self._human_speed.update({'1': 'normal speed', 
                                  str(1.):'normal speed',
                                  '2': 'double speed',
                                  str(2.): 'double speed',
                                  '3': 'triple speed',
                                  str(3.): 'triple speed',
                                  '-1': 'sdrawkcab speed',
                                  str(-1.): 'sdrawkcab speed',
                                  '10': 'flash speed',
                                  str(10.): 'flash speed',
                                  '-10': '(: deesp live',
                                  str(-10.): '(: deesp live'})

        self._background = rectangle.Rectangle(size=size, position=(0,0))
        self._background.ordering = widget.NEAR
        self._bg_set = False

        pgmrender = render.get_renderer()
        self._background.aspect_style = pgmrender.LAYOUT_FILLED
        
        self._background.opacity = 200

        self._volume = 0
        self._speed = 0
        self._control_menu = None
        
        group.Group.__init__(self)

        self.size = size
        self.add(self._background)

        # status label
        self._status_label = label.Label(text="", size=(size[0], 50),
                                         font_size=24)
        self._status_label.yx_align = 1.
        self._status_label.ordering = widget.NEAR
        self._status_label.x = 10
        self._status_label.opacity = 255
        self._status_label.tz = 1
        self.add(self._status_label)

        # progress label
        self._progress_label = label.Label(text="", size=(350,50),
                                           font_size=23)        
        self._progress_label.x_align = 1.
        self._progress_label.ordering = widget.NEAR
        self._progress_label.position = (size[0]-370, 70)
        self._progress_label.opacity = 255
        self._progress_label.tz = 1
        self.add(self._progress_label)

        # settings label
        self._settings_label = label.Label(text="", size=(450,50),
                                           font_size=23)        
        self._settings_label.x_align = 0.
        self._settings_label.ordering = widget.NEAR
        self._settings_label.position = (10, 70)
        self._settings_label.opacity = 255
        self._settings_label.tz = 1
        self.add(self._settings_label)

        # progress bar
        bg_color = color.Color((30,30,30,255))
        fg_color = color.Color((255, 255, 255, 255))
        self._progress_bar = progress_bar.ProgressBar((size[0]-30, 30),
                                                      fg_color, bg_color, 255)
        self.add(self._progress_bar)
        self._progress_bar.position = (10, 45)
        
        settings_signal.connect(self._settings_changed)
        playing_signal.connect(self._playing)
        state_changed.connect(self._state_changed)
        
        self._timeout = 0
        self._ctrl = None

    def get_skin(self):
        return self._skin

    def _application_started(self):
        appli = common.get_application()
        self._skin = appli.get_skin()

    def _key_pressed(self, key_event):
        key = key_event.keyval
        pgmrender = render.get_renderer()
        if self._control_menu and self._control_menu.visible and not self._control_menu.is_locked():
            if key == pgmrender.KEY_LEFT:
                self._control_menu.move_left()
            elif key == pgmrender.KEY_RIGHT:
                self._control_menu.move_right()
            elif key == pgmrender.KEY_RETURN:
                self._control_menu.activate_current()
            if self._timed_out and self._timed_out.active():
                self._timed_out.reset(self.hide_timeout)
                self._control_menu.unlock_animation()
                
    def get_window(self):
        return self.get_skin().get_window()

    def _state_changed(self, obj, prop, new_value):
        """
        """
        if prop == 'player-stats':
            if len(new_value) == 3:
                label, progress, fraction = new_value
                self._status_label.text = label
                self._progress_label.text = progress
                self._progress_bar.fraction = fraction
        elif isinstance(obj, menu.MenuItem) and self._control_menu:
            widget = self._control_menu.get_widget_item_for_menu_item(obj)
            if widget:
                menu_item = obj
                if prop == 'icon-path':
                    picture_path = self.get_skin().get_picture_path(new_value)
                    img = image.Image(picture_path)
                    widget.set_background_image(img)
                elif prop == 'action-callback':
                    menu_item.set_action_callback(new_value)
                elif prop == 'icon-label':
                    menu_item.set_short_name(new_value)
                
    def _settings_changed(self, setting, new_value):
        if setting == 'volume':
            self._volume =  new_value
        elif setting == 'playback-speed':
            self._speed = self._get_human_speed(new_value)
        self.update_settings()

    def update_settings(self):
        vol_str = _('Volume: %s/100') % self._volume
        self._settings_label.text = "%s, %s" % (vol_str, self._speed)

    def _get_human_speed(self, speed):
        human_speed = self._human_speed.get(str(speed), 'speed: %sX' % speed)
        return human_speed
    
    def _playing(self, playable, start):
        player = common.get_application().get_player()
        speed = player.get_playback_speed()
        self._speed = self._get_human_speed(speed)
        self._volume = player.get_volume()
        self.update_settings()

    def menuitem__get(self):
        player = common.get_application().get_player()
        playable = player.get_current_item()
        playable_uri = playable.get_uri()
        menu_manager = self.get_skin().get_menu_manager()
        menu_item = menu_manager.get_menu_item_from_uri(playable_uri)
        return menu_item

    def plugin__get(self):
        player = common.get_application().get_player()
        playable = player.get_current_item()
        playable_uri = playable.get_uri()
        plugin_manager = common.get_plugin_manager()
        if playable.contains_video():
            if str(playable_uri).startswith('dvd://'):
                name = 'dvd'
            else:
                name = 'movies'
        else:
            name = 'music'
        plugin = plugin_manager.get_plugin_with_name(name)
        return plugin
        
    def controls__get(self):
        item = self.menuitem
        controls = []
        if item:
            self.plugin.item_action_menu(item)
            controls = self.menuitem._items
        return controls
    
    def size__set(self, size):
        group.Group.size__set(self, size)
        if self._background:
            self._background.size__set(size)
    
    def width__set(self, width):
        group.Group.width__set(self, width)
        self._progress_bar.width = width
        if self._background:
            self._background.width = width
        if self._status_label:
            self._status_label.width = width
        if self._control_menu:
            self._control_menu.width = width
            
    def height__set(self, height):
        group.Group.height__set(self, height)
        if self._background:
            self._background.height = height

    def timeout__set(self, timeout):
        self._timeout = timeout
            
    def timeout__get(self):
        return self._timeout
    
    def visible__set(self, visible):
                
        # FIXME
        if visible:
            opacity = 255
            self.y = 0
        else:
            opacity = 0

        self._status_label.opacity = opacity
        self._progress_label.opacity = opacity
        self._settings_label.opacity = opacity

        if not self._bg_set:
            if self.get_skin():
                self._background.media = image.Image(self.get_skin().get_picture_path('../dock_background.png'))
                self._bg_set = True
            
        group.Group.visible__set(self, visible)
            
    def hide(self, animate=False):
        self._timed_out = None
        def _hide(c):
            self.visible = False
            self.hide_controls()
        if animate:
            self.animate(cb=_hide)
        else:
            self.visible = False
            self.hide_controls()

    def show(self, animate=True, autohide=True, show_controls=True):
        def _show(c):
            if show_controls:
                self.show_controls()
            self.visible = True
        if animate:
            self.animate(autohide=autohide, cb=_show)
        else:
            if show_controls:
                self.show_controls()
            self.visible = True

        if autohide:
            self._timed_out = reactor.callLater(self.hide_timeout, self.hide,
                                                True)

    def animate(self, autohide=False, cb=None):
        visible = not self.visible
        self.timeout = 1
        if self.timeout:
            win_height = self.get_window().canvas_size[1]

            if visible:
                start = -self.height
                end = 0
            else:
                start = 0
                end = -self.height

            duration = self.timeout * 1000
            timeline_top = [ keyframe.KeyFrame(start, 0),
                             keyframe.KeyFrame(end, 1) ]

            if not cb:
                cb = self._dfr_toggle

            prop = 'y'
            mod_top = modifier.Modifier([self,], prop, timeline_top)

            modifiers = [mod_top]

            # controls menu animation
            if self._control_menu:
                if visible:
                    start_b = win_height
                    end_b = win_height - (self._control_menu.height * 2)
                else:
                    start_b = win_height - (self._control_menu.height * 2)
                    end_b = win_height

                timeline_bottom = [ keyframe.KeyFrame(start_b, 0),
                                    keyframe.KeyFrame(end_b, 1) ]

                mod_bottom = modifier.Modifier([self._control_menu,], prop, timeline_bottom)

                modifiers.append(mod_bottom)

            self._ctrl = controller.Controller(begin=0, duration=duration,
                                               resolution=2,
                                               modifiers=modifiers,
                                               end_behavior=controller.END,
                                               interpolation=controller.SINUSOIDAL,
                                               end_callback=cb)
            if self.get_skin():
                self._ctrl.start()

    def _dfr_toggle(self, ctrl):

        self._ctrl = None
            
    def toggle(self, ctrl):
        self._ctrl = None
        self.timeout = 0
        if self.visible:
            self.hide()
        else:
            self.show()
        
    def hide_controls(self):
        if self._control_menu:
            self._control_menu.visible = False
            win = self.get_skin().get_window()
            win.sig_key_press_event.disconnect(self._key_pressed)

    def show_controls(self):
        if self.get_skin():
            win = self.get_skin().get_window()
            
            if not self._control_menu:
                self._control_menu = ControlMenu(self, self.menuitem)
                self._control_menu.update()
                win.add(self._control_menu)
                
            def connect():
                win.sig_key_press_event.connect(self._key_pressed)
                    
            reactor.callLater(0.1, connect)

            self._control_menu.visible = True
            self._control_menu.y = win.canvas_size[1] - 100
            self._control_menu.ordering = widget.NEAR

    def _player_stopped(self, playable):
        if self._control_menu:
            self._control_menu.visible = False
            self.remove(self._control_menu)
            self._control_menu = None
