# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.base_plugins import DataAccessPlugin, DataAccessFile
from elisa.core.interfaces import IDataAccess
from elisa.core import media_uri
import os, re
try:
    import gnomevfs
except:
    "GnomeVFS module not found"
    
import mimetypes

class VFSPlugin(DataAccessPlugin):

    __implements__ = IDataAccess
    __platforms__ = ['posix']

    name = 'vfs'

    provides = ["uri:%s" % protocol for protocol in ("smb", "sftp","http",
                                                     "ftp",) ]

    hidden_file_pattern = re.compile("\..*")
    
    singleton = True
    uncacheable_uri_schemes = ['smb', 'http']
    
    def __init__(self):
        DataAccessPlugin.__init__(self)
        self.cwd = gnomevfs.URI(os.getcwd())
        self.smb_groups = []
        
    def _get_gvfs_uri(self, uri):
        if type(uri) == unicode:
            uri = uri.encode('utf-8')
        uri = media_uri.Uri(location=str(uri))
        uri = self.cwd.resolve_relative(str(uri))
        return uri

    def _uri_join(self, uri, filename):
        uri = media_uri.Uri(location=str(uri))
        if uri.scheme == 'smb':
            if uri.path == '/':
                # we're browsing SMB workgroups. GnomeVFS doesn't support
                # smb://workgroup/host .. so we remove the workgroup
                # name from the uri
                f_uri = 'smb://' + filename
                if filename not in self.smb_groups:
                    self.smb_groups.append(f_uri)
            else:
                f_uri = str(uri.join(filename))
                in_workgroup = self._uri_in_group(f_uri)
                if in_workgroup:
                    f_uri = 'smb://%s' % filename
        else:
            f_uri = str(uri.join(filename))
        return f_uri

    def _uri_in_group(self, f_uri):
        return len(filter(lambda group_uri: f_uri.startswith(group_uri),
                          self.smb_groups)) > 0

    def create_menu_for_uri(self, uri, for_plugin, short_name=None):
        uri = str(self._get_gvfs_uri(uri))

        if short_name:
            if self.hidden_file_pattern.match(short_name):
                return None

        return DataAccessPlugin.create_menu_for_uri(self, uri, for_plugin,
                                                    short_name=short_name)

    def create_child_menuitems_for_uri(self, uri, start_item, nb_items,
                                       for_plugin, short_name, menu_item=None):
        uri = self._get_gvfs_uri(uri)
        return DataAccessPlugin.create_child_menuitems_for_uri(self, uri,
                                                               start_item,
                                                               nb_items,
                                                               for_plugin,
                                                               short_name,
                                                               menu_item=menu_item)

    def is_dir(self, uri):
        if str(uri).find('m3u') != -1:
            return False
        uri = str(self._get_gvfs_uri(uri))
        try:
            info = gnomevfs.get_file_info(uri).type
        except Exception, exc:
            print exc
            return False
        if info == gnomevfs.FILE_TYPE_DIRECTORY:
            return True
        elif info == gnomevfs.FILE_TYPE_REGULAR:
            return False
        elif info == gnomevfs.FILE_TYPE_SYMBOLIC_LINK:
            try:
                gnomevfs.open_directory(uri)
                return True
            except:
                return False
            
        return False
    
    def open_directory(self, uri):
        uri = self._get_gvfs_uri(uri)
        if not str(uri).endswith('/'):
            uri = self._get_gvfs_uri("%s/" % str(uri))
        try:
            handle = gnomevfs.open_directory(uri)
        except Exception, exc:
            self.logger.info("Problem accessing %s : %s" % (uri,exc))
            self.logger.info("Exception type: %s" % exc.__class__)
            handle = None
        return handle


    def get_file_type(self, uri):
        """
        Return the type of file. If directory return empty string, either
        return the content-type of the uri
        """
        if self.is_dir(uri):
            file_type = ''
        else:
            file_type, sub_type = mimetypes.guess_type(uri)
            if not file_type:
                # use tag browser
                # FIXME
                #file_type = self.tag_browser.get_content_type(uri)
                file_type = 'audio'
            else:
                parts = file_type.split('/')
                file_type = parts[0]
                if file_type == 'application':
                    # use tag browser
                    # FIXME
                    #file_type = self.tag_browser.get_content_type(uri)
                    file_type = 'audio'
        return file_type

    def get_direct_children(self, uri):
        uri = self._get_gvfs_uri(uri)
        children = {}
        if self.is_dir(uri):
            handle = self.open_directory(uri)
            index = 0
            for i in handle:
                if self.hidden_file_pattern.match(i.name):
                    continue

                
                child_uri = self._uri_join(uri, i.name)
                child_type = 'file'
                if self.is_dir(child_uri):
                    child_type = 'directory'
                file_type = self.get_file_type(child_uri)
                children[str(child_uri)] = (child_type, file_type, i.name,
                                            index)
                index += 1
        return children

    def next_location(self, from_uri, item_filter=None, root=None):
        """
        Return the uri just next to `from_uri`.
        """
        from_uri = uri = self.cwd.resolve_relative(str(from_uri))

        if not self.is_dir(uri):
            uri = uri.parent

        i = 0
        first = None
        current_found = False

        children = self.get_direct_children(uri)
        files = [ name for (u,(typ,m_type, name)) in children.iteritems() ]
        files.sort()

        try:
            index = files.index(from_uri.short_name) +1
        except ValueError:
            index = 0

        try:
            short_name = files[index]
            next_uri = str(uri) + '/' + short_name
        except IndexError:
            next_uri = None
            short_name = None

        if next_uri and root:
            self._add_to_history(root, next_uri, short_name)

        return (next_uri, short_name)

    def open(self, uri, mode=None):
        """
        Open an uri and return DataAccessFile file
        """
        file = self._get_gvfs_uri(uri)

        modes = { None: gnomevfs.OPEN_READ,
                  'r': gnomevfs.OPEN_READ,
                  'a+': gnomevfs.OPEN_WRITE
                  }
        mode = modes.get(mode)
        try:
            handle = gnomevfs.open(file, mode)
        except Exception, exc:
            #self.logger.info("Couln't open %s : %s" % (file, exc))
            handle = None
        else:
            handle = DataAccessFile(self, handle)
        return handle

    def remove_location(self, uri):
        """
        Delete data at location identified by its uri.
        """
        self.logger.info("Removing %s" % str(uri))
        #gnomevfs.unlink(str(uri))


    def read(self, handle, size=-1):
        data = ""
        try:
            if size == -1:
                remaining_size = handle.get_file_handle().get_file_info().size
                while remaining_size > 0:
                    data_chunk = handle.get_file_handle().read(remaining_size)
                    data += data_chunk
                    remaining_size -= len(data_chunk) 
            else:
                data = handle.get_file_handle().read(size)
        except gnomevfs.EOFError, error:
            print error
        return data
