# -*- coding: utf-8 -*-
# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright (C) 2006,2007, Elisa team <contact@fluendo.com>

from elisa.core.interfaces import IMediaPlayer
from elisa.core.base_plugins import PlayerPlugin, TreePlugin
from elisa.core.menu import MenuItem
from elisa.core.utils.misc import handle_failure
from elisa.core import common
import elisa.core.utils.misc
from elisa.core.utils.misc import get_resource_path

from twisted.internet import threads

import time
import os
import shutil

import sys

try:
    if sys.hexversion >= 0x20500f0:
        from xml.etree import cElementTree
    else:
        import cElementTree
except ImportError:
    cElementTree = None

class WebRadioTreePlugin(PlayerPlugin):
    """
    web radios plugin_tree
    """
    __implements__ = IMediaPlayer

    singleton = True
    name = 'web_radio'
    menu_name = _('Radios')
    
    default_config = {
        'stations':['file://./sample_data/radio/stations_big.xml',],
        'playlists': ['file://./sample_data/radio/playlists/',]
    }
    
    parent = 'music'
    provides = ['audio',]
    
    item_icon_path = 'playlist.png'

    #item_filter = lambda x,y: file_is_playlist(y)

    db_tables = {'radio': """
    id INTEGER PRIMARY KEY,
    name TEXT NOT NULL,
    urls TEXT NOT NULL,
    genre TEXT NOT NULL,
    rank INT NOT NULL"""}

    folder_action = None

    def __init__(self):
        PlayerPlugin.__init__(self)
        path = 'radios.png'
        self.set_icon_path(path)
        self.alpha = [chr(i) for i in range(65,91)]
        self.new_radios_count = 0
        self.media_mgr = common.get_application().get_media_manager()
        
    def add_wizard_shortcuts(self):
        pass

    def post_init(self):
        if self.media_mgr.enabled():
            PlayerPlugin.post_init(self)
            dfr = threads.deferToThread(self.load_stations)
            dfr.addCallback(self.create_menu)
            dfr.addErrback(handle_failure)
        elif not cElementTree:
            self.logger.info("cElementTree not found, disabling plugin")
            self.unload()
        else:
            self.unload()

        # copies sample stations list in configuration directory on first run
        appli = common.get_application()
        if appli.first_run:
            src_path = get_resource_path(self.__module__,
                                         'data/stations_big.xml')
            dest_path = appli.get_config().get_config_dir()
            shutil.copy(src_path, dest_path)
            # don't use os.path.join. media_uri separator is /
            stations_path = ["file://%s/stations_big.xml" % dest_path]
            appli.get_config().set_option("stations", stations_path,
                                          "plugins.web_radio")

    def load_stations(self):
        """
        Load all radio stations from XML and lazily insert them in DB
        """
        locations = self.get_config().get('stations',[])
        if type(locations) != list:
            locations = [locations,]
        files_not_found = []
        appli = common.get_application()
        skin = appli.get_skin()
        if skin and appli.first_run:
            msg = _("Initializing web_radios stations. It can take a while")
            skin.notify(msg)

        for location in locations:
            if not self.media_mgr._is_running():
                break
            data_plugin = self.get_data_plugin(location)
            if not data_plugin:
                continue
            handle = data_plugin.open(location)
            if not handle:
                files_not_found.append(location)
                continue
            data = data_plugin.read(handle)
            tree = cElementTree.fromstring(data)
            for radio in tree.findall('Radio'):
                if not self.media_mgr._is_running():
                    break
                name = radio.find('Name').text
                urls = radio.find('Server').text
                genre = radio.find('Genre').text
                rating = radio.find('Rank').text
                radio_id = int(radio.find('ID').text)
                db_radio = self.get_radio_with_id(radio_id)
                if not db_radio:
                    self.add_radio(radio_id, name, urls, genre, rating)
                    time.sleep(0.01)
            data_plugin.close(handle)
        if files_not_found and skin:
            msg = _("%(stations_count)s stations files not found:" %
                    {'stations_count': len(files_not_found)})
            msg = "%s\n%s" % (msg, "\n".join(files_not_found))
            #print msg
            print "f: ", files_not_found
            skin.notify(msg)
            

    def create_menu(self, r):
        self.add_item(self.create_name_folder())
        for name in ("genre", "rating"):
            item = self.create_folder_item("By %s" % name, self._get_by, name)
            self.add_item(item)
        if self.new_radios_count:
            skin = common.get_application().get_skin()
            if skin:
                skin.notify(_("%(radios_count)s new web radios found") %
                            {'radios_count': self.new_radios_count})
            self.new_radios_count = 0
            
    def create_radio_item(self, name, urls):
        icon_path = 'radio.png'
        item = MenuItem(short_name=name)
        item.set_icon_path(icon_path)
        # XXX: we take only the first url
        url = urls.split(';')[0]
        item.set_target_path(url)
        item.set_action_callback(self.play_item)
        self.item_action_menu(item)
        return item
    
    def create_folder_item(self, name, loadmore_callback, target_path):
        folder = 'folder-radios.png'
        item = MenuItem(short_name=name)
        item.set_icon_path(folder)
        item.set_target_path(target_path)
        item.set_loadmore_callback(loadmore_callback)
        item.set_items_count(1)
        return item

    def _build_names_ranges(self):
        alpha = [chr(i) for i in range(65,91)]
        result = []
        index = 0
        while index < len(alpha):
            i = 0
            sub = []
            while i < 4 and index < len(alpha):
                letter = alpha[index]
                sub.append(letter)
                i += 1
                index += 1
            result.append(sub)
        return result

    def create_name_folder(self):
        folder = 'folder-radios.png'
        item = MenuItem(short_name='By name')
        item.set_icon_path(folder)
        item.set_target_path('name')
        item.set_loadmore_callback(self._get_by_name)
        item.set_items_count(1)
        return item

    def _get_by_name(self, target_path, start_item, nb_items, menu_item):
        items = []
        targets = []
        cb = getattr(self, '_get_for_%s' % target_path)

        for alpha_range in self._build_names_ranges():
            tp = ','.join(alpha_range)
            if alpha_range not in targets:
                label = '%s-%s' % (alpha_range[0], alpha_range[-1])
                item = self.create_folder_item(label, cb, tp)
                targets.append(alpha_range)
                items.append(item)

        return items
    
    def _get_by(self, target_path, start_item, nb_items, menu_item):
        cb = getattr(self, '_get_for_%s' % target_path)
        targets = []
        items = []
        
        for radio in self.get_radios():
            genre = radio.genre
            rating = radio.rank
            
            target = locals()[target_path]
            if type(target) == int:
                target = str(target)
            else:
                target = elisa.core.utils.misc.fs_encode(target)
                
            if target not in targets:
                item = self.create_folder_item(target, cb, target)
                targets.append(target)
                items.append(item)

        if target_path == 'genre':
            items.sort(key=lambda item: item.get_short_name())
        else:
            items.sort(key=lambda item: int(item.get_short_name()))
        return items
    
    def _get_for_genre(self, genre_target, start_item, nb_items, menu_item):
        items = []
        for radio in self.get_radios_for_genre(genre_target):
            item = self.create_radio_item(radio.name, radio.urls)
            items.append(item)
        items.sort(key=lambda item: item.get_short_name())
        return items
    
    def _get_for_rating(self, rating_target, start_item, nb_items, menu_item):
        items = []
        for radio in self.get_radios_for_rating(rating_target):
            item = self.create_radio_item(radio.name, radio.urls)
            items.append(item)
        items.sort(key=lambda item: item.get_short_name())
        return items
    
    def _get_for_name(self, name_target, start_item, nb_items, menu_item):
        items = []
        for radio in self.get_radios_starting_with(name_target):
            item = self.create_radio_item(radio.name, radio.urls)
            items.append(item)
                            
        items.sort(key=lambda item: item.get_short_name())
        return items


    def item_action_menu(self, menu_item):
        PlayerPlugin.item_action_menu(self, menu_item)

        for target in ('speed+', 'speed-', 'seek+', 'seek-'):
            menu_item.del_item_with_target(target)

        item_targets = ('remove','stop','play/pause', 'vol-','vol+')
        items = [ menu_item.get_item_with_target(t) for t in item_targets ]

        for target in item_targets:
            menu_item.del_item_with_target(target)
        for item in items:
            menu_item.add_item(item)
        
        player = self.get_player()
        if player and str(menu_item.get_target_path()) == \
               str(player.get_playing_uri()):
            self.playing(menu_item)

    def play_item(self, menu_item):
        player = self.get_player()
        music_file = menu_item.get_target_path()
        
        self.logger.info('Playing radio %s' % music_file)

        def launch(r):
            player.play()
            PlayerPlugin.play_item(self, menu_item)
        
        dfr = player.set_uri(music_file)
        dfr.addCallback(launch)
        
    item_action = play_item

    def remove_parent_item(self, menu_item):
        parent = menu_item.get_parent()
        location = parent.get_target_path()
        db_radio = self.get_radio_with_urls(location)
        if db_radio:
            name = elisa.core.utils.misc.fs_encode(db_radio.name)
            self.logger.info("Removed radio %s" % name)
            self.del_radio_with_id(db_radio.id)

    # SQL API

    def get_radios(self):
        return self.sql_execute("select * from radio")

    def get_radio_with_urls(self, urls):
        radio = None
        if type(urls) == unicode:
            urls = urls.encode('utf-8')
        result = self.sql_execute("select * from radio where urls=%r" % urls)
        if result:
            radio = result[0]
        return radio

    def add_radio(self, radio_id, name, urls, genre, rating):
        name = repr(elisa.core.utils.misc.fs_encode(name))
        urls = repr(elisa.core.utils.misc.fs_encode(urls))
        genre = repr(elisa.core.utils.misc.fs_encode(genre))
        rating = int(rating)
        req = "insert into radio(id, name,urls,genre,rank) values(%s, %s, %s, %s, %s)"
        req = req % (radio_id, name, urls, genre, rating)
        self.sql_execute(req)
        self.new_radios_count += 1
        
    def del_radio_with_id(self, radio_id):
        self.sql_execute("delete from radio where id=%s" % radio_id)
        
    def get_radio_with_id(self, radio_id):
        radios = self.sql_execute("select * from radio where id=%s limit 1" % radio_id)
        radio = None
        if radios:
            radio = radios[0]
        return radio

    def get_radios_for_genre(self, genre):
        return self.sql_execute("select * from radio where genre=%s" % repr(elisa.core.utils.misc.fs_encode(genre)))

    def get_radios_for_rating(self, rating):
        return self.sql_execute("select * from radio where rank=%s" % int(rating))

    def get_radios_starting_with(self, letters):
        criteria = 'or '.join(["name like '%s%%'" % r for r in letters])
        return self.sql_execute("select * from radio where %s" % criteria) 
