# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from elisa.core import menu
from twisted.trial import unittest
import sets

class TestMenuTree(unittest.TestCase):

    args = (0, -1, lambda x: False)


    def test_parent(self):
        """ MenuTrees have no parent and are self-rooted
        """
        root = menu.MenuTree()
        self.assertEqual(root.get_parent(), None)
        self.assertEqual(root.get_root(), root)

    def test_items(self):
        """ MenuTrees store MenuItem instances
        """

        # empty root
        root = menu.MenuTree()

        # prepare a menu item
        item = menu.MenuItem("item1")
        self.assertEqual(item.get_parent(), None)

        # add the item in the tree
        root.add_item(item)
        self.assertEqual(len(root._items), 1)
        self.assertEqual(item.get_parent(), root)

        # remove item from tree
        root.del_item(item)
        self.assertEqual(len(root._items), 0)
        self.assertEqual(item.get_parent(), None)

    def test_as_item(self):
        """ MenuTrees can be converted to MenuItems
        """

        root = menu.MenuTree()
        item = menu.MenuItem("item1")
        root.add_item(item)

        root_item = root.as_menu_item()
        self.assertEqual(root_item.get_short_name(),
                         root.get_short_name())

        def got_items(r):

            def root_items(root_menu_items):
                self.assertEqual(root_menu_items, r)
                root_items.called = True
            root_items.called = False
            
            dfr = root_item.get_items(*self.args)
            dfr.addCallback(root_items)
            dfr.addCallback(lambda _: self.assertEqual(root_items.called, True))
            return dfr

        dfr = root.get_items(*self.args)
        dfr.addCallback(got_items)
        return dfr

class TestMenuItem(unittest.TestCase):

    def test_enable(self):
        """ Test menu item enabling/disabling
        """
        item = menu.MenuItem("foo")
        self.assertEqual(item.is_enabled(), True)
        item.disable()
        self.assertEqual(item.is_enabled(), False)

    def test_root(self):
        """ Test menu item / root linking
        """
        item = menu.MenuItem("foo")

        self.assertEqual(item.get_root(), None)

        # store the item in a tree and check its root
        root = menu.MenuTree()
        root.add_item(item)
        self.assertEqual(item.get_root(), root)

        # going one level further
        item2 = menu.MenuItem("bar")
        item.add_item(item2)
        self.assertEqual(item2.get_root(), root)

    def test_level(self):
        """ Test depth levels of menus
        """
        root = menu.MenuTree()
        item = menu.MenuItem("foo")
        item2 = menu.MenuItem("bar")

        item.add_item(item2)
        root.add_item(item)
        self.assertEqual(item2.get_level(), 2)

        item.del_item(item2)
        self.assertEqual(item2.get_level(), 0)

    def test_callbacks(self):
        """ Test wether menu item callback system works
        """
        item = menu.MenuItem("foo")

        for cb in item.callbacks.values():
            self.assertEqual(cb, None)

        s1 = sets.Set(['loadmore','action'])
        s2 = sets.Set(item.callbacks.keys())
        self.assert_(s1.issubset(s2))

        item.set_action_callback(self.action)

        item.fire_action('foo','bar')
        self.assertEqual(self.action_args, ('foo','bar'))

    def action(self, *args):
        self.action_args = args
