# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006,2007 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

from pgm.timing import controller, modifier, keyframe
from pgm.layout import group
import widget_item
import mutex
from elisa.core import log

WAIT = 1
NORMAL = 2

class WidgetLevel(group.Group):
    
    logger = log.Logger()
    
    def __init__(self, arrow_image=None, visible_item_count=5, width=800):
        self._added_witem = []
        group.Group.__init__(self)
        #FIXME temporay variables
        self._visible_item_count = visible_item_count
        self._active_item_position = (self._visible_item_count + 1) / 2 - 1
        self._active_item_index = 0
        self._added_widget_item = 0
        self._space = 5
        self._status= WAIT
        self._position_offset = (0,0)
        self._unselected_opacity = 150
        self.width = width
        self._animation_mutex = mutex.mutex()
        self._arrow_image = None
        self._first_draw_level_done = False
        self._set_arrow_image(arrow_image)
        self._level_loaded = False

    def width__set(self, width):
        group.Group.width__set(self,width)
        self._item_width = (self._width - (self._space * \
                                           (self._visible_item_count - 1))) \
                                           / self._visible_item_count 
        self._item_height = self._item_width
        
    
    def get_status(self):
        return self._status
    
    def set_status(self, status):
        self._status = status
        
    def lock_animation(self):
        return self._animation_mutex.testandset()
    
    def is_locked(self):    
        return self._animation_mutex.locked
    
    def unlock_animation(self):
        self._animation_mutex.unlock()
    
    def _set_arrow_image(self, arrow_image):
        self._arrow_image = arrow_image
    
    def get_arrow_image(self):
        return self._arrow_image
        
    def __repr__(self):
        return "<%s %s children>" % (self.__class__.__name__,
                                     self.get_widget_count())
   
    def get_selected_rank(self):
        return self._active_item_index 

    def rotate_image(self, uri, angle):
        pass

    def first_draw(self):
        if self._status == NORMAL and not self._first_draw_level_done:
            self.first_draw_level()
        
    def first_draw_level(self):
        widget_count = self.get_widget_count()
        if widget_count > 0:
            self._first_draw_level_done = True
            rank = 0
            offset = self._visible_item_count - self._active_item_position -1
            for witem in self.get_widget_items(0, self._visible_item_count + 1):
                self._draw_widget_item(witem, rank-offset)
                rank += 1
            
            item_count = rank
            self._active_item_index = self._active_item_position
            #ajust item position for put selected item at center
            #calculate the number of move_left needed for having the selected item at self._active_item_index
            if item_count < self._visible_item_count:
                _black_item = self._visible_item_count - item_count 
                _move_left = ( _black_item + 1 ) / 2
                self._active_item_index -= _move_left
                offset = ( self._item_width + self._space ) * _move_left
                for witem in self.get_visible_widget_items():
                    if witem:
                        witem.x += offset
            selected_item = self.get_widget_item(self._active_item_index)
            if selected_item:
                selected_item.opacity = 255
            self._update_active_list()
    
    def show_selected_item_arrow(self):
        selected_item = self.get_widget_item(self._active_item_index)
        if selected_item:
            selected_item.show_label()
            selected_item.show_arrow()
    
    def hide_selected_item_arrow(self):
        selected_item = self.get_widget_item(self._active_item_index)
        if selected_item:
            selected_item.hide_arrow()
            selected_item.hide_label()
        
    def get_x_from_index(self, index):
        x = ( self._item_width + self._space ) * self.get_relative_index(index)
        return x
    
    def get_relative_index(self, absolute_index):
        item_on_each_side = self._visible_item_count - (self._active_item_position + 1) +1
        offscreen_left_index = self._active_item_index - item_on_each_side
        relative_index = absolute_index - offscreen_left_index - 1
        return relative_index
        
    def get_widget_items(self, start , count):
        return self._added_witem[start:start+count]
    
    def get_widget_item(self, index, create_in_not_exist=False):
        #create_in_not_exist=False is used for children classes cache system integration
        size = len(self._added_witem)
        if index < 0 or index > size - 1:
            return None
        return self._added_witem[index]
    
    def is_visible(self, index):
        if self._first_draw_level_done:
            if index >= self.get_first_visible_item_index() and index <= self.get_last_visible_item_index():
                return True
        elif index >=0 and index < self._visible_item_count + 2:
            return True
        return False
        
    def get_widget_count(self):
        return self._added_widget_item
                
    def add_item(self, picture):
        size = (self._item_width, self._item_width)
        witem = widget_item.WidgetItem(picture, start_size=size,
                                       arrow_image=self.get_arrow_image())
        self._added_widget_item += 1
        self._added_witem.append(witem)
        witem.opacity = self._unselected_opacity
        return witem
    
    def get_first_visible_item_index(self):
        #in fact, the returned rank is the first (and the only) offscreen widget at left
        item_on_each_side = self._visible_item_count - (self._active_item_position + 1) +1
        offscreen_left_index = self._active_item_index - item_on_each_side
        return offscreen_left_index
    
    def get_last_visible_item_index(self):
        #in fact, the returned rank is the last (and the only) offscreen widget at right
        item_on_each_side = self._visible_item_count - (self._active_item_position + 1) +1
        offscreen_right_index = self._active_item_index + item_on_each_side
        return offscreen_right_index
    
    def get_visible_widget_items(self):
        offscreen_left_index = self.get_first_visible_item_index()
        offscreen_right_index = self.get_last_visible_item_index()
        count = offscreen_right_index - max(0,offscreen_left_index) + 1
        return self.get_widget_items(max(0,offscreen_left_index), count)
    
    def remove_widget_item(self, witem):
        witem.delete()
        self.remove(witem)

    def _draw_widget_item(self, witem, index):
        x = self.get_x_from_index(index)
        witem.position = (x, 0)
        self.add( witem )

    def move_left(self):
        if self.lock_animation(): 
            if self._active_item_index > 0:
                witem = self.get_widget_item(self._active_item_index)
                if witem:
                    witem.hide_arrow()
                    witem.hide_label()
                    self._anim_in_progress = 'left'
                    self._horiz_anim('left')
            else:
                self.unlock_animation()


    def _horiz_anim(self, direction):
        max = self._item_width + self._space
        selected_item = self.get_widget_item(self._active_item_index)
        if selected_item:
            selected_item.opacity = self._unselected_opacity
            if direction == 'left':
                end_frame = keyframe.KeyFrame(max, 1.0)
            else:
                end_frame = keyframe.KeyFrame(-max, 1.0)
            timeline_pos = [keyframe.KeyFrame(0, 0.0), end_frame]
            mdf_pos = modifier.Modifier(self.get_visible_widget_items(), 'x',
                                        timeline_pos, 
                                        operation=modifier.RELATIVE)
            ctrl = controller.Controller(begin=0, duration=250,
                                         resolution=15,
                                         end_behavior=controller.END,
                                         modifiers=[mdf_pos],
                                         end_callback=self._anim_end)
            ctrl.start()

        
    def move_right(self): 
        if self.lock_animation():
            if self._active_item_index < self.get_widget_count()-1:
                witem = self.get_widget_item(self._active_item_index)
                witem.hide_arrow()
                witem.hide_label()
                self._anim_in_progress = 'right'
                self._horiz_anim('right')
            else:
                self.unlock_animation()    
                    
    def _anim_end(self, ctrl):
        if self._anim_in_progress == 'left':
            #remove of window last widget
            self._active_item_index -=1
        else:
            self._active_item_index +=1    
            
        self.anim_end(self._anim_in_progress)
        self._update_active_list()
        witem = self.get_widget_item(self._active_item_index)
        if witem:
            witem.opacity = 255
        self.unlock_animation()
        self._anim_in_progress = None
        
    def anim_end(self, anim_label):
        pass
                  
    def _update_active_list(self):
        #remove offscreen item, and add futur onscreen item
        item_on_each_side = self._visible_item_count - (self._active_item_position + 1) +1
        offscreen_left_index = self._active_item_index - item_on_each_side
        offscreen_right_index = self._active_item_index + item_on_each_side
        
        #check if widget offscreen exists and create it
        if offscreen_left_index >= 0:
            widget_left_offscreen = self.get_widget_item(offscreen_left_index, True)
            if widget_left_offscreen and widget_left_offscreen not in self.children:
                self.add(widget_left_offscreen)
                widget_left_offscreen.opacity = self._unselected_opacity 
                widget_left_offscreen.position = (self.get_x_from_index(offscreen_left_index), self.y)
        
        if offscreen_right_index < self.get_widget_count():
            widget_right_offscreen = self.get_widget_item(offscreen_right_index, True)
            if widget_right_offscreen and widget_right_offscreen not in self.children:
                self.add(widget_right_offscreen)
                widget_right_offscreen.opacity = self._unselected_opacity 
                widget_right_offscreen.position = (self.get_x_from_index(offscreen_right_index), self.y)
                
        #remove widget offscreen with index > ffscreen_right_index
        #or index < offscreen_left_index
        if offscreen_left_index > 0:
            widget_left_offscreen = self.get_widget_item(offscreen_left_index - 1, False)
            if widget_left_offscreen and widget_left_offscreen in self.children:
                self.remove(widget_left_offscreen)
        
        if offscreen_right_index + 1 < self.get_widget_count():
            widget_right_offscreen = self.get_widget_item(offscreen_right_index + 1, False)
            if widget_right_offscreen and widget_right_offscreen in self.children:
                self.remove(widget_right_offscreen)
                
        
if __name__ == '__main__':  
    global wl
    from pgm import render, widget
    from pgm.layout import window
    from pgm.form import rectangle, image
    import sys, os
    
    try:
        from twisted.internet import glib2reactor
        glib2reactor.install()
    except AssertionError:
        # already installed...
        pass
    
    from twisted.internet import reactor, task
    from twisted.spread import pb
    import gobject
    
    def exit_event_cb(event):    
        win.quit_loop()
        reactor.stop()
    
    def _get_translate_modifier(self, widget_list, start_x, end_x, operation, property='x'):
        timeline_pos = [ keyframe.KeyFrame(start_x, 0.0), keyframe.KeyFrame(end_x, 1.0) ]
        return modifier.Modifier(widget_list, property,
                                            timeline_pos, 
                                            operation=operation)
    
    def key_press_cb(event):
        if event.keyval == 22:
            wl.move_right()
        elif event.keyval == 20:
            wl.move_left()
        elif event.keyval == 15: #space bar
            start_anim()
        elif event.keyval == render.get_renderer().KEY_q:
            reactor.stop()
            
    render.init('opengl1')
    title = 'widget_level test'
    global win
    win = window.Window(title=title)
    
    
    """
    #pigment re-order test
    fxd = fixed.Fixed()
    win.add(fxd)
    _background_widget = rectangle.Rectangle((800,600))
    fxd.put(_background_widget,(0,0))
    img = image.Image('pictures/background.png')
    _background_widget.set_media(img)
    """

    base = 'elisa/core/plugins/skins/default/'
    wl = WidgetLevel(image.Image(os.path.join(base,'pictures/arrow.png')))
    #wl.y = 150
    
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/pictures.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/dvd.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/movies.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/music.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/config.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/play.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/pause.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/dvd.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/music.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/music.png')))
    wl.add_item(image.Image(os.path.join(base,'pictures/256X256/remove.png')))
    
    wl.set_status(NORMAL)
    win.add(wl)
    
    wl.first_draw()
    
    wl.show_selected_item_arrow()
    
    _background_widget = rectangle.Rectangle((800,600))
    img = image.Image(os.path.join(base,'pictures/background.png'))
    _background_widget.media = img
    _background_widget.ordering = widget.FAR
    win.add(_background_widget)
    
    win.sig_key_press_event.connect(key_press_cb)
    win.sig_exit_event.connect(exit_event_cb)
    gobject.timeout_add(15, win.update)

    win.main(reactor.loop)
 
    reactor.run()
    
    
